import asyncio
import random
import time
import sys
import pytest
from threading import Thread

from bugsnag import BreadcrumbType, Breadcrumb, Breadcrumbs
from bugsnag.utils import FilterDict


def test_breadcrumb_types():
    assert len(BreadcrumbType) == 8

    assert BreadcrumbType.NAVIGATION.value == 'navigation'
    assert BreadcrumbType.REQUEST.value == 'request'
    assert BreadcrumbType.PROCESS.value == 'process'
    assert BreadcrumbType.LOG.value == 'log'
    assert BreadcrumbType.USER.value == 'user'
    assert BreadcrumbType.STATE.value == 'state'
    assert BreadcrumbType.ERROR.value == 'error'
    assert BreadcrumbType.MANUAL.value == 'manual'


def test_breadcrumb_properties():
    breadcrumb = Breadcrumb(
        'This is a test breadcrumb',
        BreadcrumbType.REQUEST,
        {'a': 1, 'b': 2, 'c': 3, 'xyz': 456},
        '1234-56-78T12:34:56.789+00:00'
    )

    assert breadcrumb.message == 'This is a test breadcrumb'
    assert breadcrumb.type == BreadcrumbType.REQUEST
    assert breadcrumb.metadata == {'a': 1, 'b': 2, 'c': 3, 'xyz': 456}
    assert breadcrumb.timestamp == '1234-56-78T12:34:56.789+00:00'


def test_breadcrumb_properties_with_keyword_params():
    breadcrumb = Breadcrumb(
        timestamp='9876-54-32T12:34:56.789+00:00',
        metadata={'a': 1, 'b': 2, 'c': 3, 'xyz': 456},
        type=BreadcrumbType.PROCESS,
        message='This is a second test breadcrumb'
    )

    assert breadcrumb.message == 'This is a second test breadcrumb'
    assert breadcrumb.type == BreadcrumbType.PROCESS
    assert breadcrumb.metadata == {'a': 1, 'b': 2, 'c': 3, 'xyz': 456}
    assert breadcrumb.timestamp == '9876-54-32T12:34:56.789+00:00'


def test_breadcrumb_to_dict():
    breadcrumb = Breadcrumb(
        'This is another test breadcrumb',
        BreadcrumbType.ERROR,
        {'abc': 123, 'xyz': 'fourfivesix'},
        '1234-56-78T12:34:56.789+00:00'
    )

    breadcrumb_dict = breadcrumb.to_dict()

    expected = {
        'name': 'This is another test breadcrumb',
        'type': BreadcrumbType.ERROR.value,
        'metaData': FilterDict({'abc': 123, 'xyz': 'fourfivesix'}),
        'timestamp': '1234-56-78T12:34:56.789+00:00'
    }

    assert breadcrumb_dict == expected

    # pytest allows the previous assertion to pass when metadata is not a
    # FilterDict, so explicitly check for this as it's necessary for redaction
    assert isinstance(breadcrumb_dict['metaData'], FilterDict)


def test_there_is_a_max_number_of_breadcrumbs():
    breadcrumbs = Breadcrumbs(max_breadcrumbs=2)

    first_breadcrumb = Breadcrumb('hello', BreadcrumbType.ERROR, {}, 'time')
    second_breadcrumb = Breadcrumb('hi', BreadcrumbType.REQUEST, {}, 'emit')

    breadcrumbs.append(first_breadcrumb)
    breadcrumbs.append(second_breadcrumb)

    breadcrumb_list = breadcrumbs.to_list()

    assert len(breadcrumb_list) == 2
    assert breadcrumb_list[0] == first_breadcrumb
    assert breadcrumb_list[1] == second_breadcrumb

    third_breadcrumb = Breadcrumb('howdy', BreadcrumbType.LOG, {}, 'now')
    breadcrumbs.append(third_breadcrumb)

    breadcrumb_list = breadcrumbs.to_list()

    # the length should match the maximum number of breadcrumbs set earlier
    assert len(breadcrumb_list) == 2
    assert breadcrumb_list[0] == second_breadcrumb
    assert breadcrumb_list[1] == third_breadcrumb


def test_the_number_of_breadcrumbs_can_be_reduced():
    breadcrumbs = Breadcrumbs(max_breadcrumbs=3)

    first_breadcrumb = Breadcrumb('hello', BreadcrumbType.ERROR, {}, 'time')
    second_breadcrumb = Breadcrumb('hi', BreadcrumbType.REQUEST, {}, 'emit')
    third_breadcrumb = Breadcrumb('howdy', BreadcrumbType.LOG, {}, 'now')

    breadcrumbs.append(first_breadcrumb)
    breadcrumbs.append(second_breadcrumb)
    breadcrumbs.append(third_breadcrumb)

    breadcrumb_list = breadcrumbs.to_list()

    assert len(breadcrumb_list) == 3
    assert breadcrumb_list[0] == first_breadcrumb
    assert breadcrumb_list[1] == second_breadcrumb
    assert breadcrumb_list[2] == third_breadcrumb

    breadcrumbs.resize(1)
    breadcrumb_list = breadcrumbs.to_list()

    assert len(breadcrumb_list) == 1
    assert breadcrumb_list[0] == third_breadcrumb

    breadcrumbs.resize(0)
    breadcrumb_list = breadcrumbs.to_list()

    assert len(breadcrumb_list) == 0


def test_the_number_of_breadcrumbs_can_be_increased():
    breadcrumbs = Breadcrumbs(max_breadcrumbs=1)

    first_breadcrumb = Breadcrumb('hello', BreadcrumbType.ERROR, {}, 'time')

    breadcrumbs.append(first_breadcrumb)

    breadcrumb_list = breadcrumbs.to_list()

    assert len(breadcrumb_list) == 1
    assert breadcrumb_list[0] == first_breadcrumb

    second_breadcrumb = Breadcrumb('hi', BreadcrumbType.REQUEST, {}, 'emit')
    third_breadcrumb = Breadcrumb('howdy', BreadcrumbType.LOG, {}, 'now')

    breadcrumbs.resize(5)

    breadcrumbs.append(second_breadcrumb)
    breadcrumbs.append(third_breadcrumb)

    breadcrumb_list = breadcrumbs.to_list()

    assert len(breadcrumb_list) == 3
    assert breadcrumb_list[0] == first_breadcrumb
    assert breadcrumb_list[1] == second_breadcrumb
    assert breadcrumb_list[2] == third_breadcrumb

    breadcrumbs.append(second_breadcrumb)
    breadcrumbs.append(third_breadcrumb)

    breadcrumb_list = breadcrumbs.to_list()

    assert len(breadcrumb_list) == 5
    assert breadcrumb_list[0] == first_breadcrumb
    assert breadcrumb_list[1] == second_breadcrumb
    assert breadcrumb_list[2] == third_breadcrumb
    assert breadcrumb_list[3] == second_breadcrumb
    assert breadcrumb_list[4] == third_breadcrumb


def test_the_breadcrumb_list_is_separate_on_different_threads():
    breadcrumbs = Breadcrumbs(max_breadcrumbs=5)

    def append_breadcrumbs(id):
        try:
            thread.exception = None

            first_breadcrumb = Breadcrumb(
                'a' + id,
                BreadcrumbType.ERROR,
                {'a': id, 'b': 'xyz'},
                'x'
            )

            second_breadcrumb = Breadcrumb(
                'b' + id,
                BreadcrumbType.USER,
                {'b': id, 'c': 'xyz'},
                'y'
            )

            third_breadcrumb = Breadcrumb(
                'c' + id,
                BreadcrumbType.STATE,
                {'c': id, 'xyz': 'yes'},
                'z'
            )

            breadcrumbs.append(first_breadcrumb)
            breadcrumbs.append(second_breadcrumb)
            breadcrumbs.append(third_breadcrumb)

            # sleep for a bit to allow other threads time to interfere
            time.sleep(random.randrange(0, 100) / 1000)

            breadcrumb_list = breadcrumbs.to_list()

            assert len(breadcrumb_list) == 3
            assert breadcrumb_list[0] == first_breadcrumb
            assert breadcrumb_list[1] == second_breadcrumb
            assert breadcrumb_list[2] == third_breadcrumb

            breadcrumbs.append(second_breadcrumb)
            breadcrumbs.append(third_breadcrumb)

            # sleep for a bit to allow other threads time to interfere
            time.sleep(random.randrange(0, 100) / 1000)

            breadcrumb_list = breadcrumbs.to_list()

            assert len(breadcrumb_list) == 5
            assert breadcrumb_list[0] == first_breadcrumb
            assert breadcrumb_list[1] == second_breadcrumb
            assert breadcrumb_list[2] == third_breadcrumb
            assert breadcrumb_list[3] == second_breadcrumb
            assert breadcrumb_list[4] == third_breadcrumb

            breadcrumbs.resize(2)

            # sleep for a bit to allow other threads time to interfere
            time.sleep(random.randrange(0, 100) / 1000)

            breadcrumb_list = breadcrumbs.to_list()

            assert len(breadcrumb_list) == 2
            assert breadcrumb_list[0] == second_breadcrumb
            assert breadcrumb_list[1] == third_breadcrumb

        except Exception as e:
            thread.exception = e

    threads = []
    for i in range(5):
        thread = Thread(target=append_breadcrumbs, args=[str(i)])
        threads.append(thread)

    # shuffle the list of threads so they don't run in a reliable order
    random.shuffle(threads)

    for thread in threads:
        thread.start()

    # shuffle the list of threads so they don't run in a reliable order
    random.shuffle(threads)

    for thread in threads:
        thread.join(2)

        assert not thread.is_alive()

        # if an exception happened in the thread, raise it here instead
        if thread.exception is not None:
            raise thread.exception


@pytest.mark.skipif(
    sys.version_info < (3, 7),
    reason="requires ContextVar support (Python 3.7 or higher)"
)
def test_the_breadcrumb_list_is_separate_on_different_async_contexts():
    breadcrumbs = Breadcrumbs(max_breadcrumbs=5)

    async def append_breadcrumbs(id):
        # create a copy of the breadcrumbs for this context
        breadcrumbs.create_copy_for_context()

        first_breadcrumb = Breadcrumb(
            'a' + id,
            BreadcrumbType.ERROR,
            {'a': id, 'b': 'xyz'},
            'x'
        )

        second_breadcrumb = Breadcrumb(
            'b' + id,
            BreadcrumbType.USER,
            {'b': id, 'c': 'xyz'},
            'y'
        )

        third_breadcrumb = Breadcrumb(
            'c' + id,
            BreadcrumbType.STATE,
            {'c': id, 'xyz': 'yes'},
            'z'
        )

        breadcrumbs.append(first_breadcrumb)
        breadcrumbs.append(second_breadcrumb)
        breadcrumbs.append(third_breadcrumb)

        await asyncio.sleep(random.randrange(0, 100) / 1000)

        breadcrumb_list = breadcrumbs.to_list()

        assert len(breadcrumb_list) == 3
        assert breadcrumb_list[0] == first_breadcrumb
        assert breadcrumb_list[1] == second_breadcrumb
        assert breadcrumb_list[2] == third_breadcrumb

        breadcrumbs.append(second_breadcrumb)
        breadcrumbs.append(third_breadcrumb)

        await asyncio.sleep(random.randrange(0, 100) / 1000)

        breadcrumb_list = breadcrumbs.to_list()

        assert len(breadcrumb_list) == 5
        assert breadcrumb_list[0] == first_breadcrumb
        assert breadcrumb_list[1] == second_breadcrumb
        assert breadcrumb_list[2] == third_breadcrumb
        assert breadcrumb_list[3] == second_breadcrumb
        assert breadcrumb_list[4] == third_breadcrumb

        breadcrumbs.resize(2)

        await asyncio.sleep(random.randrange(0, 100) / 1000)

        breadcrumb_list = breadcrumbs.to_list()

        assert len(breadcrumb_list) == 2
        assert breadcrumb_list[0] == second_breadcrumb
        assert breadcrumb_list[1] == third_breadcrumb

    async def test():
        tasks = []
        for i in range(5):
            tasks.append(asyncio.ensure_future(append_breadcrumbs(str(i))))

        await asyncio.gather(*tasks)

    loop = asyncio.new_event_loop()

    try:
        loop.run_until_complete(test())
    finally:
        loop.close()


def test_the_breadcrumb_list_can_be_cleared():
    breadcrumbs = Breadcrumbs(max_breadcrumbs=5)

    first_breadcrumb = Breadcrumb('hello', BreadcrumbType.ERROR, {}, 'time')
    second_breadcrumb = Breadcrumb('hi', BreadcrumbType.REQUEST, {}, 'emit')
    third_breadcrumb = Breadcrumb('howdy', BreadcrumbType.LOG, {}, 'now')

    breadcrumbs.append(first_breadcrumb)
    breadcrumbs.append(second_breadcrumb)
    breadcrumbs.append(third_breadcrumb)

    breadcrumb_list = breadcrumbs.to_list()

    assert len(breadcrumb_list) == 3
    assert breadcrumb_list[0] == first_breadcrumb
    assert breadcrumb_list[1] == second_breadcrumb
    assert breadcrumb_list[2] == third_breadcrumb

    breadcrumbs.clear()

    assert len(breadcrumbs.to_list()) == 0



import warnings
import sys

from bugsnag import Configuration
from bugsnag.delivery import (
    UrllibDelivery,
    RequestsDelivery,
    create_default_delivery
)

from tests.utils import IntegrationTest


class DeliveryTest(IntegrationTest):

    def setUp(self):
        super(DeliveryTest, self).setUp()
        self.config = Configuration()
        self.config.configure(api_key='abc',
                              endpoint=self.server.url,
                              asynchronous=False)

    def test_urllib_delivery(self):
        UrllibDelivery().deliver(self.config, '{"legit": 4}')

        self.assertSentReportCount(1)
        request = self.server.received[0]
        self.assertEqual(request['json_body'], {"legit": 4})
        self.assertEqual(request['headers']['Content-Type'],
                         'application/json')

    def test_create_default_delivery(self):
        delivery = create_default_delivery()

        if 'requests' in sys.modules:
            self.assertTrue(isinstance(delivery, RequestsDelivery))
        else:
            self.assertTrue(isinstance(delivery, UrllibDelivery))

    def test_misconfigured_sessions_endpoint_sends_warning(self):
        delivery = create_default_delivery()

        with warnings.catch_warnings(record=True) as warn:
            warnings.simplefilter("always")
            delivery.deliver_sessions(self.config, '{"apiKey":"aaab"}')
            self.assertEqual(1, len(warn))
            self.assertEqual(0, len(self.server.received))
            self.assertTrue('No sessions will be sent' in str(warn[0].message))
            delivery.deliver_sessions(self.config, '{"apiKey":"aaab"}')
            self.assertEqual(1, len(warn))
            self.assertEqual(0, len(self.server.received))

        self.config.configure(session_endpoint=self.server.url)

        with warnings.catch_warnings(record=True) as warn:
            warnings.simplefilter("always")
            delivery.deliver_sessions(self.config, '{"apiKey":"aaab"}')
            self.assertEqual(0, len(warn))
            self.assertEqual(1, len(self.server.received))



from collections import OrderedDict
from typing import Dict, List, Union


class FeatureFlag:
    def __init__(
        self,
        name: Union[str, bytes],
        variant: Union[None, str, bytes] = None
    ):
        self._name = name
        self._variant = self._coerce_variant(variant)

    @property
    def name(self) -> Union[str, bytes]:
        return self._name

    @property
    def variant(self) -> Union[None, str, bytes]:
        return self._variant

    # for JSON encoding the feature flag
    def to_dict(self) -> Dict[str, Union[str, bytes]]:
        if self._variant is None:
            return {'featureFlag': self._name}

        return {'featureFlag': self._name, 'variant': self._variant}

    # a FeatureFlag is valid if it has a non-empty string name and a variant
    # that's None or a string
    # FeatureFlags that are not valid will be ignored
    def is_valid(self) -> bool:
        return (
            isinstance(self._name, (str, bytes)) and
            len(self._name) > 0 and
            (self._variant is None or isinstance(self._variant, (str, bytes)))
        )

    def __eq__(self, other) -> bool:
        if isinstance(other, FeatureFlag):
            return (
                self._name == other._name and
                self._variant == other._variant
            )

        return NotImplemented

    def __hash__(self):
        return hash((self._name, self._variant))

    def __repr__(self):
        return 'FeatureFlag(%s, %s)' % (repr(self._name), repr(self._variant))

    def _coerce_variant(
        self,
        variant: Union[None, str, bytes]
    ) -> Union[None, str, bytes]:
        if variant is None or isinstance(variant, (str, bytes)):
            return variant

        try:
            return str(variant)
        except Exception:
            return None


class FeatureFlagDelegate:
    def __init__(self):
        self._storage = OrderedDict()

    def add(
        self,
        name: Union[str, bytes],
        variant: Union[None, str, bytes]
    ) -> None:
        flag = FeatureFlag(name, variant)

        if flag.is_valid():
            self._storage[flag.name] = flag

    def merge(self, flags: List[FeatureFlag]) -> None:
        for flag in flags:
            if isinstance(flag, FeatureFlag) and flag.is_valid():
                self._storage[flag.name] = flag

    def remove(self, name: Union[str, bytes]) -> None:
        if name in self._storage:
            del self._storage[name]

    def clear(self) -> None:
        self._storage.clear()

    def copy(self) -> 'FeatureFlagDelegate':
        copy = FeatureFlagDelegate()
        copy._storage = self._storage.copy()

        return copy

    def to_list(self) -> List[FeatureFlag]:
        return list(self._storage.values())

    def to_json(self) -> List[Dict[str, Union[str, bytes]]]:
        return [flag.to_dict() for flag in self.to_list()]



import unittest

import bugsnag
from bugsnag.middleware import MiddlewareStack, skip_bugsnag_middleware
from bugsnag.configuration import RequestConfiguration


class SampleMiddlewareClass(object):
    def __init__(self, callback):
        self.callback = callback
        self.char = None

    def __call__(self, item):
        item.append(self.char)
        self.callback(item)


class SampleMiddlewareClassA(SampleMiddlewareClass):
    def __init__(self, callback):
        SampleMiddlewareClass.__init__(self, callback)
        self.char = 'A'


class SampleMiddlewareClassB(SampleMiddlewareClass):
    def __init__(self, callback):
        SampleMiddlewareClass.__init__(self, callback)
        self.char = 'B'


class SampleMiddlewareClassC(SampleMiddlewareClass):
    def __init__(self, callback):
        SampleMiddlewareClass.__init__(self, callback)
        self.char = 'C'


class SampleMiddlewareReturning(object):
    def __init__(self, callback):
        pass

    def __call__(self, item):
        return


def create_event(exception=None) -> bugsnag.Event:
    return bugsnag.Event(
        exception or RuntimeError('oh no!'),
        bugsnag.configure(),
        RequestConfiguration.get_instance()
    )


class TestMiddleware(unittest.TestCase):

    def test_order_of_middleware(self):

        a = []

        m = MiddlewareStack()

        m.before_notify(lambda _: a.append(1))
        m.before_notify(lambda _: a.append(2))

        m.after_notify(lambda _: a.append(4))
        m.after_notify(lambda _: a.append(3))

        m.run(create_event(), lambda: None)

        self.assertEqual(a, [1, 2, 3, 4])

    def test_before_notify_returning_false(self):

        a = []

        m = MiddlewareStack()

        m.before_notify(lambda _: False)
        m.before_notify(lambda _: a.append(1))

        m.run(create_event(), lambda: a.append(2))

        self.assertEqual(a, [])

    def test_before_exception_handling(self):

        a = []

        m = MiddlewareStack()
        m.before_notify(lambda _: a.penned(1))
        m.run(create_event(), lambda: a.append(2))

        self.assertEqual(a, [2])

    def test_after_exception_handling(self):
        a = []

        m = MiddlewareStack()
        m.after_notify(lambda _: a.penned(1))
        m.run(create_event(), lambda: a.append(2))

        self.assertEqual(a, [2])

    def test_insert_before_ordering(self):
        a = []

        m = MiddlewareStack()
        m.append(SampleMiddlewareClassA)
        m.append(SampleMiddlewareClassB)
        m.insert_before(SampleMiddlewareClassA, SampleMiddlewareClassC)
        m.run(a, lambda: a.append('Callback'))

        self.assertEqual(a, ['C', 'A', 'B', 'Callback'])

    def test_insert_before_default_ordering(self):
        a = []

        m = MiddlewareStack()
        m.append(SampleMiddlewareClassA)
        m.append(SampleMiddlewareClassB)
        m.insert_before(SampleMiddlewareClass, SampleMiddlewareClassC)
        m.run(a, lambda: a.append('Callback'))

        self.assertEqual(a, ['A', 'B', 'C', 'Callback'])

    def test_insert_after_ordering(self):
        a = []

        m = MiddlewareStack()
        m.append(SampleMiddlewareClassA)
        m.append(SampleMiddlewareClassB)
        m.insert_after(SampleMiddlewareClassA, SampleMiddlewareClassC)
        m.run(a, lambda: a.append('Callback'))

        self.assertEqual(a, ['A', 'C', 'B', 'Callback'])

    def test_insert_after_default_ordering(self):
        a = []

        m = MiddlewareStack()
        m.append(SampleMiddlewareClassA)
        m.append(SampleMiddlewareClassB)
        m.insert_after(SampleMiddlewareClass, SampleMiddlewareClassC)
        m.run(a, lambda: a.append('Callback'))

        self.assertEqual(a, ['A', 'B', 'C', 'Callback'])

    def test_callback_not_run_if_middleware_returns(self):
        a = []

        m = MiddlewareStack()
        m.append(SampleMiddlewareClassA)
        m.append(SampleMiddlewareReturning)
        m.append(SampleMiddlewareClassB)
        m.run(a, lambda: a.append('Callback'))

        self.assertEqual(a, ['A'])

    def test_skip_bugsnag_middleware_returns_false_when_attr_is_present(self):
        exception = Exception('oh no')
        exception.skip_bugsnag = True

        event = create_event(exception)

        assert skip_bugsnag_middleware(event) is False

    def test_skip_bugsnag_middleware_returns_none_when_attr_is_missing(self):
        exception = Exception('oh no')
        event = create_event(exception)

        assert skip_bugsnag_middleware(event) is None



from typing import Any, Dict, Optional, List, Union  # noqa
import linecache
import logging
import os
import sys
import traceback
import inspect
import warnings
from copy import deepcopy

import bugsnag

from bugsnag.breadcrumbs import Breadcrumb
from bugsnag.notifier import _NOTIFIER_INFORMATION
from bugsnag.utils import (
    fully_qualified_class_name as class_name,
    FilterDict,
    SanitizingJSONEncoder
)
from bugsnag.error import Error
from bugsnag.feature_flags import FeatureFlag, FeatureFlagDelegate

__all__ = ('Event',)

if sys.version_info < (3, 11):
    try:
        from exceptiongroup import BaseExceptionGroup
    except ImportError:
        # we're on Python < 3.11 and exceptiongroup isn't installed
        # an empty tuple can be passed to 'isinstance' safely and will
        # always return false, so we default to that
        BaseExceptionGroup = ()


class Event:
    """
    An occurrence of an exception for delivery to Bugsnag
    """
    NOTIFIER_NAME = _NOTIFIER_INFORMATION['name']
    NOTIFIER_URL = _NOTIFIER_INFORMATION['url']
    PAYLOAD_VERSION = "4.0"
    SUPPORTED_SEVERITIES = ["info", "warning", "error"]

    def __init__(self, exception: BaseException, config, request_config,
                 **options):
        """
        Create a new event

        exception is the exception being reported.
        config is the global instance of bugsnag.Configuration
        request_config is the thread-local instance of bugsnag.Configuration
        (used by middleware)

        options can be used to override any of the configuration parameters:
            "api_key", "release_stage", "app_version", "hostname"
        and to provide the following top-level event payload keys:
            "user", "context", "severity", "grouping_hash", "metadata",
            ("user_id")
        or to provide the exception parameter:
            "traceback"
        All other keys will be sent as metadata to Bugsnag.
        """
        self._exception = exception
        self._original_error = exception

        self.options = options
        self.config = config
        self.request_config = request_config
        self.request = None  # type: Any
        self._breadcrumbs = [
            deepcopy(breadcrumb) for breadcrumb in config.breadcrumbs
        ]
        self._feature_flag_delegate = options.pop(
            'feature_flag_delegate',
            FeatureFlagDelegate()
        ).copy()

        def get_config(key):
            return options.pop(key, getattr(self.config, key))

        self.release_stage = get_config("release_stage")
        self.app_version = get_config("app_version")
        self.app_type = get_config("app_type")
        self.hostname = get_config("hostname")
        self.runtime_versions = get_config("runtime_versions")
        self.send_code = get_config("send_code")

        self.context = options.pop("context", None)
        self.severity = options.pop("severity", "warning")
        if self.severity not in self.SUPPORTED_SEVERITIES:
            self.severity = "warning"

        self.unhandled = options.pop("unhandled", False)
        self.severity_reason = options.pop(
            "severity_reason",
            {'type': 'handledException'}
        )

        self.user = options.pop("user", {})
        if "user_id" in options:
            self.user["id"] = options.pop("user_id")

        # for backwards compatibility we generate the first error's stacktrace
        # here and use it as 'self.stacktrace' and 'self.errors[0].stacktrace'
        # this allows mutations 'self.stacktrace' to be reflected in the errors
        # list, which is used to generate the JSON payload
        stacktrace = self._generate_stacktrace(
            self.options.pop(
                "traceback",
                getattr(exception, '__traceback__', sys.exc_info()[2])
            ),
            self.options.pop("source_func", None)
        )

        self._stacktrace = stacktrace
        self._errors = self._generate_error_list(exception, stacktrace)

        self.grouping_hash = options.pop("grouping_hash", None)
        self.api_key = options.pop("api_key", get_config("api_key"))

        self.session = None  # type: Optional[Dict]

        self.metadata = {}  # type: Dict[str, Dict[str, Any]]
        if 'meta_data' in options:
            warnings.warn('The Event "meta_data" argument has been replaced ' +
                          'with "metadata"', DeprecationWarning)
            for name, tab in options.pop("meta_data").items():
                self.add_tab(name, tab)

        for name, tab in options.pop('metadata', {}).items():
            self.add_tab(name, tab)

        for name, tab in options.items():
            self.add_tab(name, tab)

        if hasattr(exception, "__notes__"):
            self.add_tab(
                "exception notes",
                dict(enumerate(exception.__notes__)) # type: ignore # noqa
            )

    @property
    def meta_data(self) -> Dict[str, Dict[str, Any]]:
        warnings.warn('The Event "meta_data" property has been replaced ' +
                      'with "metadata".', DeprecationWarning)
        return self.metadata

    @property
    def breadcrumbs(self) -> List[Breadcrumb]:
        return self._breadcrumbs.copy()

    @property
    def errors(self) -> List[Error]:
        return self._errors.copy()

    @property
    def original_error(self) -> BaseException:
        return self._original_error

    @property
    def stacktrace(self) -> List[Dict[str, Any]]:
        warnings.warn(
            (
                'The Event "stacktrace" property has been deprecated in favour'
                ' of accessing the stacktrace of an error, for example '
                '"errors[0].stacktrace"'
            ),
            DeprecationWarning
        )

        return self._stacktrace

    @stacktrace.setter
    def stacktrace(self, value: List[Dict[str, Any]]) -> None:
        warnings.warn(
            (
                'The Event "stacktrace" property has been deprecated in favour'
                ' of accessing the stacktrace of an error, for example '
                '"errors[0].stacktrace"'
            ),
            DeprecationWarning
        )

        self._stacktrace = value
        self._errors[0].stacktrace = value

    @property
    def exception(self) -> BaseException:
        warnings.warn(
            (
                'The Event "exception" property has been replaced with '
                '"original_error"'
            ),
            DeprecationWarning
        )

        return self._exception

    @exception.setter
    def exception(self, value: BaseException) -> None:
        warnings.warn(
            (
                'Setting the Event "exception" property has been deprecated, '
                'update the "errors" list instead'
            ),
            DeprecationWarning
        )

        self._exception = value
        self._errors[0] = Error(
            class_name(value),
            str(value),
            self._stacktrace
        )

    def set_user(self, id=None, name=None, email=None):
        """
        Set user parameters on event.
        """
        if id:
            self.user["id"] = id
        if name:
            self.user["name"] = name
        if email:
            self.user["email"] = email

    def add_custom_data(self, key, value):
        """
        Add data to the "custom" metadata tab
        """
        self.add_tab("custom", {key: value})

    def add_tab(self, name, dictionary):
        """
        Add a metadata tab to the event

        If the tab already exists, the new content will be merged into the
        existing content.
        """
        if not isinstance(dictionary, dict):
            self.add_tab("custom", {name: dictionary})
            return

        if name not in self.metadata:
            self.metadata[name] = {}

        self.metadata[name].update(dictionary)

    @property
    def feature_flags(self) -> List[FeatureFlag]:
        return self._feature_flag_delegate.to_list()

    def add_feature_flag(
        self,
        name: Union[str, bytes],
        variant: Union[None, str, bytes] = None
    ) -> None:
        self._feature_flag_delegate.add(name, variant)

    def add_feature_flags(self, feature_flags: List[FeatureFlag]) -> None:
        self._feature_flag_delegate.merge(feature_flags)

    def clear_feature_flag(self, name: Union[str, bytes]) -> None:
        self._feature_flag_delegate.remove(name)

    def clear_feature_flags(self) -> None:
        self._feature_flag_delegate.clear()

    def _generate_error_list(
        self,
        exception: BaseException,
        first_error_stacktrace: List[Dict[str, Any]]
    ) -> List[Error]:
        error_list = [
            Error(
                class_name(exception),
                str(exception),
                first_error_stacktrace
            )
        ]

        if not isinstance(exception, BaseException):
            return error_list

        while True:
            if exception.__cause__:
                exception = exception.__cause__
            elif exception.__context__ and not exception.__suppress_context__:
                exception = exception.__context__
            else:
                break

            error_list.append(
                Error(
                    class_name(exception),
                    str(exception),
                    self._generate_stacktrace(exception.__traceback__)
                )
            )

        # unwrap BaseExceptionGroups so that their contained exceptions are
        # also reported
        # we don't recurse into nested BaseExceptionGroups or cause/context
        # here because there's a big risk of that leading to a huge number of
        # exceptions, which is difficult to reason about
        if isinstance(self._original_error, BaseExceptionGroup):
            for sub_exception in self._original_error.exceptions: # type: ignore # noqa
                error_list.append(
                    Error(
                        class_name(sub_exception),
                        str(sub_exception),
                        self._generate_stacktrace(sub_exception.__traceback__)
                    )
                )

        return error_list

    def _generate_stacktrace(
        self,
        tb,
        source_func=None
    ) -> List[Dict[str, Any]]:
        """
        Build the stacktrace
        """
        if tb:
            trace = traceback.extract_tb(tb)
        else:
            trace = traceback.extract_stack()

        bugsnag_module_path = os.path.dirname(bugsnag.__file__)
        logging_module_path = os.path.dirname(logging.__file__)
        exclude_module_paths = [bugsnag_module_path, logging_module_path]
        user_exclude_modules = self.config.traceback_exclude_modules
        for exclude_module in user_exclude_modules:
            try:
                module_file = exclude_module.__file__
                if module_file[-4:] == '.pyc':
                    module_file = module_file[:-1]
                exclude_module_paths.append(module_file)
            except Exception:
                self.config.logger.exception(
                    'Could not exclude module: %s' % repr(exclude_module))

        lib_root = self.config.lib_root
        if lib_root and lib_root[-1] != os.sep:
            lib_root += os.sep

        project_root = self.config.project_root
        if project_root and project_root[-1] != os.sep:
            project_root += os.sep

        stacktrace = []
        if source_func is not None:
            try:
                source = inspect.getsourcefile(source_func)
                lines = inspect.getsourcelines(source_func)
                line = 0
                if lines is not None and len(lines) > 1:
                    line = lines[1]

                trace.insert(
                    0,
                    [source, line, source_func.__name__]  # type: ignore
                )
            except (IOError, TypeError):
                pass

        for frame in trace:
            file_name = os.path.abspath(str(frame[0]))
            in_project = False

            skip_module = False
            for module_path in exclude_module_paths:
                if file_name.startswith(module_path):
                    skip_module = True
                    break
            if skip_module:
                continue

            # Fetch the code before (potentially) removing the project root
            # from the file path
            code = self._code_for(file_name, int(str(frame[1])))

            if lib_root and file_name.startswith(lib_root):
                file_name = file_name[len(lib_root):]
            elif project_root and file_name.startswith(project_root):
                file_name = file_name[len(project_root):]
                in_project = True

            stacktrace.append({
                "file": file_name,
                "lineNumber": int(str(frame[1])),
                "method": str(frame[2]),
                "inProject": in_project,
                "code": code
            })

        stacktrace.reverse()

        return stacktrace

    def _code_for(self, file_name, line, window_size=7):
        """
        Find the code around this line in the file.
        """
        if not self.send_code:
            return None

        try:
            lines = linecache.getlines(file_name)

            start = max(line - int(window_size / 2), 1)
            end = start + window_size

            # The last line of the file is len(lines). End of an
            # exclusive range is one greater.
            if end > len(lines) + 1:
                end = len(lines) + 1
                start = max(end - window_size, 1)

            return dict((n, lines[n - 1].rstrip()) for n in range(start, end))

        except Exception:
            return None

    def _payload(self):
        # Fetch the notifier version from the package
        encoder = SanitizingJSONEncoder(
            self.config.logger,
            separators=(',', ':'),
            keyword_filters=self.config.params_filters
        )

        # Construct the payload dictionary
        return encoder.encode({
            "apiKey": self.api_key,
            "notifier": _NOTIFIER_INFORMATION,
            "events": [{
                "severity": self.severity,
                "severityReason": self.severity_reason,
                "unhandled": self.unhandled,
                "releaseStage": self.release_stage,
                "app": {
                    "version": self.app_version,
                    "type": self.app_type,
                },
                "context": self.context,
                "groupingHash": self.grouping_hash,
                "exceptions": [
                    error.to_dict() for error in self.errors
                ],
                "metaData": FilterDict(self.metadata),
                "user": FilterDict(self.user),
                "device": FilterDict({
                    "hostname": self.hostname,
                    "runtimeVersions": self.runtime_versions
                }),
                "projectRoot": self.config.project_root,
                "libRoot": self.config.lib_root,
                "session": self.session,
                "breadcrumbs": [
                    breadcrumb.to_dict() for breadcrumb in self._breadcrumbs
                ],
                "featureFlags": self._feature_flag_delegate.to_json()
            }]
        })



from functools import wraps
from contextlib import contextmanager
import logging

from bugsnag.handlers import BugsnagHandler
from bugsnag.legacy import default_client
from bugsnag import Client, BreadcrumbType
import bugsnag

from tests.utils import IntegrationTest, ScaryException


def use_client_logger(func):
    @wraps(func)
    def wrapped(obj):
        client = Client(
            api_key='abcdef',
            endpoint=obj.server.url,
            asynchronous=False
        )

        handler = client.log_handler()

        with scoped_logger() as logger:
            logger.setLevel(logging.INFO)
            logger.addHandler(handler)

            func(obj, handler, logger)

    return wrapped


@contextmanager
def scoped_logger():
    logger = logging.getLogger(__name__)

    try:
        yield logger
    finally:
        logger.handlers = []
        logger.filters = []


class HandlersTest(IntegrationTest):
    def setUp(self):
        super(HandlersTest, self).setUp()
        bugsnag.configure(endpoint=self.server.url,
                          notify_release_stages=['dev'],
                          release_stage='dev',
                          asynchronous=False)
        bugsnag.logger.setLevel(logging.INFO)

    def tearDown(self):
        super(HandlersTest, self).tearDown()
        bugsnag.logger.setLevel(logging.CRITICAL)

    def test_message(self):
        handler = BugsnagHandler()
        logger = logging.getLogger(__name__)
        logger.addHandler(handler)

        logger.critical('The system is down')
        logger.removeHandler(handler)

        self.assertSentReportCount(1)

        json_body = self.server.received[0]['json_body']
        event = json_body['events'][0]
        exception = event['exceptions'][0]
        self.assertEqual('The system is down', exception['message'])

    def test_severity_critical(self):
        handler = BugsnagHandler()
        logger = logging.getLogger(__name__)
        logger.addHandler(handler)

        logger.critical('The system is down')
        logger.removeHandler(handler)

        self.assertSentReportCount(1)

        json_body = self.server.received[0]['json_body']
        event = json_body['events'][0]
        exception = event['exceptions'][0]
        self.assertEqual('LogCRITICAL', exception['errorClass'])
        self.assertEqual('error', event['severity'])
        self.assertEqual(logging.CRITICAL,
                         event['metaData']['extra data']['levelno'])
        self.assertEqual('CRITICAL',
                         event['metaData']['extra data']['levelname'])

    def test_severity_error(self):
        handler = BugsnagHandler()
        logger = logging.getLogger(__name__)
        logger.addHandler(handler)

        logger.error('The system is down')
        logger.removeHandler(handler)

        self.assertSentReportCount(1)

        json_body = self.server.received[0]['json_body']
        event = json_body['events'][0]
        exception = event['exceptions'][0]
        self.assertEqual('LogERROR', exception['errorClass'])
        self.assertEqual('error', event['severity'])
        self.assertEqual(logging.ERROR,
                         event['metaData']['extra data']['levelno'])
        self.assertEqual('ERROR',
                         event['metaData']['extra data']['levelname'])

    def test_severity_warning(self):
        handler = BugsnagHandler()
        logger = logging.getLogger(__name__)
        logger.addHandler(handler)

        logger.warning('The system is down')
        logger.removeHandler(handler)

        self.assertSentReportCount(1)
        json_body = self.server.received[0]['json_body']
        event = json_body['events'][0]
        exception = event['exceptions'][0]
        self.assertEqual('LogWARNING', exception['errorClass'])
        self.assertEqual('warning', event['severity'])
        self.assertEqual(logging.WARNING,
                         event['metaData']['extra data']['levelno'])
        self.assertEqual('WARNING',
                         event['metaData']['extra data']['levelname'])

    def test_severity_info(self):
        handler = BugsnagHandler()
        logger = logging.getLogger(__name__)
        logger.setLevel(logging.INFO)
        logger.addHandler(handler)

        logger.info('The system is down')
        logger.removeHandler(handler)

        self.assertSentReportCount(1)
        json_body = self.server.received[0]['json_body']
        event = json_body['events'][0]
        exception = event['exceptions'][0]
        self.assertEqual('LogINFO', exception['errorClass'])
        self.assertEqual('info', event['severity'])
        self.assertEqual(logging.INFO,
                         event['metaData']['extra data']['levelno'])
        self.assertEqual('INFO',
                         event['metaData']['extra data']['levelname'])

    def test_levelname_message(self):
        handler = BugsnagHandler()
        logger = logging.getLogger(__name__)
        logger.setLevel(logging.INFO)
        logger.addHandler(handler)

        class MessageFilter(logging.Filter):

            def filter(self, record):
                record.levelname = None
                return True

        handler.addFilter(MessageFilter())
        logger.info('The system is down')
        logger.removeHandler(handler)

        self.assertSentReportCount(1)
        json_body = self.server.received[0]['json_body']
        event = json_body['events'][0]
        exception = event['exceptions'][0]
        self.assertEqual('LogMessage', exception['errorClass'])

    def test_custom_level(self):
        handler = BugsnagHandler()
        logger = logging.getLogger(__name__)
        logger.addHandler(handler)

        logger.log(341, 'The system is down')
        logger.removeHandler(handler)

        self.assertSentReportCount(1)
        json_body = self.server.received[0]['json_body']
        event = json_body['events'][0]
        exception = event['exceptions'][0]
        self.assertEqual('LogLevel 341', exception['errorClass'])

    def test_custom_levelname(self):
        handler = BugsnagHandler()
        logger = logging.getLogger(__name__)
        logger.addHandler(handler)
        logging.addLevelName(402, 'OMG')

        logger.log(402, 'The system is down')
        logger.removeHandler(handler)

        self.assertSentReportCount(1)
        json_body = self.server.received[0]['json_body']
        event = json_body['events'][0]
        exception = event['exceptions'][0]
        self.assertEqual('LogOMG', exception['errorClass'])
        self.assertEqual('error', event['severity'])

    def test_exc_info(self):
        handler = BugsnagHandler()
        logger = logging.getLogger(__name__)
        logger.addHandler(handler)

        try:
            raise ScaryException('Oh no')
        except Exception:
            logger.exception('The system is down')

        logger.removeHandler(handler)

        self.assertSentReportCount(1)
        json_body = self.server.received[0]['json_body']
        event = json_body['events'][0]
        exception = event['exceptions'][0]
        self.assertEqual(exception['errorClass'], 'tests.utils.ScaryException')

    def test_extra_fields(self):
        handler = BugsnagHandler(extra_fields={'fruit': ['grapes', 'pears']})
        logger = logging.getLogger(__name__)
        logger.addHandler(handler)

        logger.error('A wild tomato appeared', extra={
            'grapes': 8, 'pears': 2, 'tomatoes': 1
        })
        logger.removeHandler(handler)

        json_body = self.server.received[0]['json_body']
        event = json_body['events'][0]
        self.assertEqual(event['metaData']['fruit'], {
            'grapes': 8, 'pears': 2
        })

    def test_client_metadata_fields(self):
        client = Client(
            api_key='abcdef',
            endpoint=self.server.url,
            asynchronous=False
        )

        handler = client.log_handler(extra_fields={
            'fruit': ['grapes', 'pears']
        })
        logger = logging.getLogger(__name__)
        logger.addHandler(handler)

        logger.error('A wild tomato appeared', extra={
            'grapes': 8, 'pears': 2, 'tomatoes': 1
        })
        logger.removeHandler(handler)

        json_body = self.server.received[0]['json_body']
        event = json_body['events'][0]
        self.assertEqual(event['metaData']['fruit'], {
            'grapes': 8, 'pears': 2
        })

    @use_client_logger
    def test_client_message(self, handler, logger):
        logger.critical('The system is down')
        self.assertSentReportCount(1)

        json_body = self.server.received[0]['json_body']
        event = json_body['events'][0]
        exception = event['exceptions'][0]

        self.assertEqual('The system is down', exception['message'])

    @use_client_logger
    def test_client_severity_critical(self, handler, logger):
        logger.critical('The system is down')

        self.assertSentReportCount(1)

        json_body = self.server.received[0]['json_body']
        event = json_body['events'][0]
        exception = event['exceptions'][0]

        self.assertEqual('LogCRITICAL', exception['errorClass'])
        self.assertEqual('error', event['severity'])
        self.assertEqual(logging.CRITICAL,
                         event['metaData']['extra data']['levelno'])
        self.assertEqual('CRITICAL',
                         event['metaData']['extra data']['levelname'])

    @use_client_logger
    def test_client_severity_error(self, handler, logger):
        logger.error('The system is down')

        self.assertSentReportCount(1)

        json_body = self.server.received[0]['json_body']
        event = json_body['events'][0]
        exception = event['exceptions'][0]

        self.assertEqual('LogERROR', exception['errorClass'])
        self.assertEqual('error', event['severity'])
        self.assertEqual(logging.ERROR,
                         event['metaData']['extra data']['levelno'])
        self.assertEqual('ERROR',
                         event['metaData']['extra data']['levelname'])

    @use_client_logger
    def test_client_severity_warning(self, handler, logger):
        logger.warning('The system is down')

        self.assertSentReportCount(1)
        json_body = self.server.received[0]['json_body']
        event = json_body['events'][0]
        exception = event['exceptions'][0]

        self.assertEqual('LogWARNING', exception['errorClass'])
        self.assertEqual('warning', event['severity'])
        self.assertEqual(logging.WARNING,
                         event['metaData']['extra data']['levelno'])
        self.assertEqual('WARNING',
                         event['metaData']['extra data']['levelname'])

    @use_client_logger
    def test_client_severity_info(self, handler, logger):
        logger.info('The system is down')

        self.assertSentReportCount(1)
        json_body = self.server.received[0]['json_body']
        event = json_body['events'][0]
        exception = event['exceptions'][0]

        self.assertEqual('LogINFO', exception['errorClass'])
        self.assertEqual('info', event['severity'])
        self.assertEqual(logging.INFO,
                         event['metaData']['extra data']['levelno'])
        self.assertEqual('INFO',
                         event['metaData']['extra data']['levelname'])

    @use_client_logger
    def test_client_add_callback(self, handler, logger):

        def some_callback(record, options):
            for key in record.meals:
                options['metadata'][key] = record.meals[key]

        handler.add_callback(some_callback)
        logger.info('Everything is fine', extra={'meals': {
            'food': {'fruit': ['pear', 'grape']},
            'drinks': {'free': 'water'}
        }})

        self.assertSentReportCount(1)
        json_body = self.server.received[0]['json_body']
        event = json_body['events'][0]
        self.assertEqual(event['metaData']['food'], {
            'fruit': ['pear', 'grape']
        })
        self.assertEqual(event['metaData']['drinks'], {
            'free': 'water'
        })

    @use_client_logger
    def test_client_remove_callback(self, handler, logger):

        def some_callback(record, options):
            options['metadata']['tab'] = {'key': 'value'}

        def some_other_callback(record, options):
            options['metadata']['tab2'] = {'key': 'value'}

        handler.add_callback(some_callback)
        handler.add_callback(some_other_callback)
        handler.remove_callback(some_callback)
        logger.info('Everything is fine')

        self.assertSentReportCount(1)
        json_body = self.server.received[0]['json_body']
        event = json_body['events'][0]
        self.assertTrue('tab' not in event['metaData'])
        self.assertEqual(event['metaData']['tab2'], {
            'key': 'value'
        })

    @use_client_logger
    def test_client_clear_callbacks(self, handler, logger):

        def some_callback(record, options):
            options['metadata']['tab'] = {'key': 'value'}

        def some_other_callback(record, options):
            options['metadata']['tab2'] = {'key': 'value'}

        handler.add_callback(some_callback)
        handler.add_callback(some_other_callback)
        handler.clear_callbacks()
        logger.info('Everything is fine')

        self.assertSentReportCount(1)
        json_body = self.server.received[0]['json_body']
        event = json_body['events'][0]
        self.assertTrue('tab' not in event['metaData'])
        self.assertTrue('tab2' not in event['metaData'])

    @use_client_logger
    def test_client_crashing_callback(self, handler, logger):

        def some_callback(record, options):
            options['metadata']['tab'] = {'key': 'value'}
            raise ScaryException('Oh dear')

        def some_other_callback(record, options):
            options['metadata']['tab']['key2'] = 'other value'

        handler.add_callback(some_callback)
        handler.add_callback(some_other_callback)
        logger.info('Everything is fine')

        self.assertSentReportCount(1)
        json_body = self.server.received[0]['json_body']
        event = json_body['events'][0]
        self.assertEqual(event['metaData']['tab'], {
            'key': 'value', 'key2': 'other value'
        })

    @use_client_logger
    def test_client_callback_exception(self, handler, logger):

        def exception_replacing_callback(record, options):
            options['exception'] = ScaryException('replacement')

        handler.add_callback(exception_replacing_callback)
        logger.info('Everything is fine')

        self.assertSentReportCount(1)
        json_body = self.server.received[0]['json_body']
        event = json_body['events'][0]
        exception = event['exceptions'][0]

        self.assertEqual(exception['errorClass'], 'tests.utils.ScaryException')
        self.assertEqual(exception['message'], 'replacement')

    @use_client_logger
    def test_client_callback_exception_metadata(self, handler, logger):

        def exception_replacing_callback(record, options):
            options['exception'] = 'metadata'

        handler.add_callback(exception_replacing_callback)
        logger.info('Everything is fine')

        self.assertSentReportCount(1)
        json_body = self.server.received[0]['json_body']
        event = json_body['events'][0]
        exception = event['exceptions'][0]

        self.assertEqual(exception['errorClass'], 'LogINFO')
        self.assertEqual(exception['message'], 'Everything is fine')
        self.assertEqual(event['metaData']['custom'],
                         {'exception': 'metadata'})

    @use_client_logger
    def test_logging_grouping_hash(self, handler, logger):
        logger.info("This happened", extra={'groupingHash': '<hash value>'})

        self.assertSentReportCount(1)
        json_body = self.server.received[0]['json_body']
        event = json_body['events'][0]
        exception = event['exceptions'][0]

        self.assertEqual(exception['message'], 'This happened')
        self.assertEqual(event['groupingHash'], '<hash value>')

    @use_client_logger
    def test_log_filter_leaves_breadcrumbs_for_logs_below_report_level(
        self,
        handler,
        logger
    ):
        logger.addFilter(handler.leave_breadcrumbs)
        handler.setLevel(logging.ERROR)

        logger.info('Everything is fine')

        assert self.sent_report_count == 0
        assert len(handler.client.configuration.breadcrumbs) == 1

        breadcrumb = handler.client.configuration.breadcrumbs[0]
        assert breadcrumb.message == 'Everything is fine'
        assert breadcrumb.type == BreadcrumbType.LOG
        assert breadcrumb.metadata == {'logLevel': 'INFO'}

        logger.error('Everything is not fine')

        assert self.sent_report_count == 1

        # we expect 2 breadcrumbs - one from the 'info' log above and one from
        # the notify caused by the 'error' log
        assert len(handler.client.configuration.breadcrumbs) == 2

        json_body = self.server.received[0]['json_body']
        event = json_body['events'][0]
        exception = event['exceptions'][0]

        assert exception['errorClass'] == 'LogERROR'
        assert exception['message'] == 'Everything is not fine'

        breadcrumb = handler.client.configuration.breadcrumbs[1]
        assert breadcrumb.message == 'LogERROR'
        assert breadcrumb.type == BreadcrumbType.ERROR
        assert breadcrumb.metadata == {
            'errorClass': 'LogERROR',
            'message': 'Everything is not fine',
            'severity': 'error',
            'unhandled': False,
        }

    @use_client_logger
    def test_log_filter_does_not_leave_breadcrumbs_for_logs_below_its_level(
        self,
        handler,
        logger
    ):
        logger.addFilter(handler.leave_breadcrumbs)
        handler.setLevel(logging.ERROR)

        handler.client.configuration.configure(
            breadcrumb_log_level=logging.WARN
        )

        logger.info('Everything is fine')

        assert self.sent_report_count == 0
        assert len(handler.client.configuration.breadcrumbs) == 0

        logger.warning('Everything might be fine')

        assert self.sent_report_count == 0
        assert len(handler.client.configuration.breadcrumbs) == 1

        breadcrumb = handler.client.configuration.breadcrumbs[0]
        assert breadcrumb.message == 'Everything might be fine'
        assert breadcrumb.type == BreadcrumbType.LOG
        assert breadcrumb.metadata == {'logLevel': 'WARNING'}

    @use_client_logger
    def test_log_filter_does_not_leave_breadcrumbs_when_bugsnag_create_breadcrumb_is_false(  # noqa: E501
        self,
        handler,
        logger
    ):
        logger.addFilter(handler.leave_breadcrumbs)
        handler.setLevel(logging.ERROR)

        logger.info(
            'Everything is fine',
            extra={'bugsnag_create_breadcrumb': False}
        )

        assert self.sent_report_count == 0
        assert len(handler.client.configuration.breadcrumbs) == 0

        logger.info(
            'Everything is fine',
            extra={'bugsnag_create_breadcrumb': True}
        )

        assert self.sent_report_count == 0
        assert len(handler.client.configuration.breadcrumbs) == 1

        breadcrumb = handler.client.configuration.breadcrumbs[0]
        assert breadcrumb.message == 'Everything is fine'
        assert breadcrumb.type == BreadcrumbType.LOG
        assert breadcrumb.metadata == {'logLevel': 'INFO'}

        logger.info('Everything is fine')

        assert self.sent_report_count == 0
        assert len(handler.client.configuration.breadcrumbs) == 2

        breadcrumb = handler.client.configuration.breadcrumbs[1]
        assert breadcrumb.message == 'Everything is fine'
        assert breadcrumb.type == BreadcrumbType.LOG
        assert breadcrumb.metadata == {'logLevel': 'INFO'}

    @use_client_logger
    def test_log_filter_does_not_leave_breadcrumbs_when_log_breadcrumbs_are_disabled(  # noqa: E501
        self,
        handler,
        logger
    ):
        logger.addFilter(handler.leave_breadcrumbs)
        handler.setLevel(logging.ERROR)
        handler.client.configuration.configure(enabled_breadcrumb_types=[])

        logger.info('Everything is fine')

        assert self.sent_report_count == 0
        assert len(handler.client.configuration.breadcrumbs) == 0

        handler.client.configuration.configure(
            enabled_breadcrumb_types=[BreadcrumbType.LOG]
        )

        logger.info('Everything is fine')

        assert self.sent_report_count == 0
        assert len(handler.client.configuration.breadcrumbs) == 1

        breadcrumb = handler.client.configuration.breadcrumbs[0]
        assert breadcrumb.message == 'Everything is fine'
        assert breadcrumb.type == BreadcrumbType.LOG
        assert breadcrumb.metadata == {'logLevel': 'INFO'}

    def test_log_filter_leaves_breadcrumbs_when_manually_constructed(self):
        default_client.configuration.max_breadcrumbs = 25

        with scoped_logger() as logger:
            handler = BugsnagHandler()
            logger.addHandler(handler)
            logger.addFilter(handler.leave_breadcrumbs)

            logger.setLevel(logging.INFO)
            handler.setLevel(logging.ERROR)

            logger.info('Everything is fine')

            assert self.sent_report_count == 0
            assert len(default_client.configuration.breadcrumbs) == 1

            breadcrumb = default_client.configuration.breadcrumbs[0]
            assert breadcrumb.message == 'Everything is fine'
            assert breadcrumb.type == BreadcrumbType.LOG
            assert breadcrumb.metadata == {'logLevel': 'INFO'}

            logger.error('Everything is not fine')

            assert self.sent_report_count == 1

            # we expect 2 breadcrumbs - one from the 'info' log above and one
            # from the notify caused by the 'error' log
            assert len(default_client.configuration.breadcrumbs) == 2

            json_body = self.server.received[0]['json_body']
            event = json_body['events'][0]
            exception = event['exceptions'][0]

            assert exception['errorClass'] == 'LogERROR'
            assert exception['message'] == 'Everything is not fine'

            breadcrumb = default_client.configuration.breadcrumbs[1]
            assert breadcrumb.message == 'LogERROR'
            assert breadcrumb.type == BreadcrumbType.ERROR
            assert breadcrumb.metadata == {
                'errorClass': 'LogERROR',
                'message': 'Everything is not fine',
                'severity': 'error',
                'unhandled': False,
            }

    def test_log_filter_leaves_breadcrumbs_when_handler_has_no_level(self):
        default_client.configuration.max_breadcrumbs = 25

        with scoped_logger() as logger:
            handler = BugsnagHandler()
            logger.addFilter(handler.leave_breadcrumbs)
            logger.setLevel(logging.DEBUG)

            logger.info('Everything is fine')

            assert self.sent_report_count == 0
            assert len(default_client.configuration.breadcrumbs) == 1

            breadcrumb = default_client.configuration.breadcrumbs[0]
            assert breadcrumb.message == 'Everything is fine'
            assert breadcrumb.type == BreadcrumbType.LOG
            assert breadcrumb.metadata == {'logLevel': 'INFO'}

            logger.error('Everything is not fine')

            assert self.sent_report_count == 0
            assert len(default_client.configuration.breadcrumbs) == 2

            breadcrumb = default_client.configuration.breadcrumbs[1]
            assert breadcrumb.message == 'Everything is not fine'
            assert breadcrumb.type == BreadcrumbType.LOG
            assert breadcrumb.metadata == {'logLevel': 'ERROR'}



import unittest
import json
import timeit
import re
import threading
import uuid
import logging
import pytest
from datetime import datetime, timedelta, timezone

from bugsnag.utils import (SanitizingJSONEncoder, FilterDict,
                           is_json_content_type, parse_content_type,
                           ThreadContextVar, to_rfc3339, remove_query_from_url)

logger = logging.getLogger(__name__)


class TestUtils(unittest.TestCase):
    def tearDown(self):
        super(TestUtils, self).tearDown()

    def test_encode_filters(self):
        data = FilterDict({"credit_card": "123213213123", "password": "456",
                           "cake": True})
        encoder = SanitizingJSONEncoder(
            logger,
            keyword_filters=["credit_card", "password"]
        )

        sane_data = json.loads(encoder.encode(data))
        self.assertEqual(sane_data, {"credit_card": "[FILTERED]",
                                     "password": "[FILTERED]",
                                     "cake": True})

    def test_encode_filters_object_key(self):
        object_key = object()
        data = FilterDict({"password": "456", object_key: True})

        encoder = SanitizingJSONEncoder(
            logger,
            keyword_filters=["password"]
        )

        actual = json.loads(encoder.encode(data))
        expected = {"password": "[FILTERED]", str(object_key): True}

        self.assertEqual(actual, expected)

    def test_encode_filters_bytes(self):
        data = FilterDict({b"credit_card": "123213213123", b"password": "456",
                           "cake": True})
        encoder = SanitizingJSONEncoder(
            logger,
            keyword_filters=["credit_card", "password"]
        )

        sane_data = json.loads(encoder.encode(data))
        self.assertEqual(sane_data, {"credit_card": "[FILTERED]",
                                     "password": "[FILTERED]",
                                     "cake": True})

    def test_sanitize_list(self):
        data = FilterDict({"list": ["carrots", "apples", "peas"],
                           "passwords": ["abc", "def"]})

        encoder = SanitizingJSONEncoder(
            logger,
            keyword_filters=["credit_card", "passwords"]
        )

        sane_data = json.loads(encoder.encode(data))
        self.assertEqual(sane_data, {"list": ["carrots", "apples", "peas"],
                                     "passwords": "[FILTERED]"})

    def test_sanitize_valid_unicode_object(self):
        data = {"item": '\U0001f62c'}
        encoder = SanitizingJSONEncoder(logger, keyword_filters=[])
        sane_data = json.loads(encoder.encode(data))
        self.assertEqual(sane_data, data)

    def test_sanitize_nested_object_filters(self):
        data = FilterDict({"metadata": {"another_password": "My password"}})

        encoder = SanitizingJSONEncoder(
            logger,
            keyword_filters=["password"]
        )

        sane_data = json.loads(encoder.encode(data))
        self.assertEqual(sane_data,
                         {"metadata": {"another_password": "[FILTERED]"}})

    def test_sanitize_bad_utf8_object(self):
        data = {"bad_utf8": "test \xe9"}
        encoder = SanitizingJSONEncoder(logger, keyword_filters=[])
        sane_data = json.loads(encoder.encode(data))
        self.assertEqual(sane_data, data)

    def test_sanitize_unencoded_object(self):
        data = {"exc": Exception()}
        encoder = SanitizingJSONEncoder(logger, keyword_filters=[])
        sane_data = json.loads(encoder.encode(data))
        self.assertEqual(sane_data, {"exc": ""})

    def test_json_encode(self):
        payload = {"a": "a" * 512 * 1024}
        expected = {"a": "a" * 1024}
        encoder = SanitizingJSONEncoder(logger, keyword_filters=[])
        self.assertEqual(json.loads(encoder.encode(payload)), expected)

    def test_filter_dict(self):
        data = FilterDict({"metadata": {"another_password": "My password"}})
        encoder = SanitizingJSONEncoder(logger, keyword_filters=["password"])
        sane_data = encoder.filter_string_values(data)
        self.assertEqual(sane_data,
                         {"metadata": {"another_password": "[FILTERED]"}})

    def test_decode_bytes(self):
        data = FilterDict({b"metadata": "value"})
        encoder = SanitizingJSONEncoder(logger, keyword_filters=["password"])
        sane_data = json.loads(encoder.encode(data))
        self.assertEqual(sane_data, {"metadata": "value"})

    def test_unfiltered_encode(self):
        data = {"metadata": {"another_password": "My password"}}
        encoder = SanitizingJSONEncoder(logger, keyword_filters=["password"])
        sane_data = json.loads(encoder.encode(data))
        self.assertEqual(sane_data, data)

    def test_thread_context_vars_get_raises_if_no_default(self):
        token = ThreadContextVar(str(uuid.uuid4()))
        self.assertRaises(LookupError, token.get)

    def test_thread_context_vars_returns_default_value_from_get(self):
        token = ThreadContextVar(str(uuid.uuid4()), default={'pips': 3})
        self.assertEqual({'pips': 3}, token.get())

    def test_thread_context_vars_set_new_value_with_no_default(self):
        token = ThreadContextVar(str(uuid.uuid4()))
        token.set({'peas': 'maybe'})
        self.assertEqual({'peas': 'maybe'}, token.get())

    def test_thread_context_vars_set_new_value(self):
        token = ThreadContextVar(str(uuid.uuid4()), default={'pips': 3})
        token.set({'carrots': 'no'})
        self.assertEqual({'carrots': 'no'}, token.get())

    def test_thread_context_vars_in_thread(self):
        """
        Verify that ThreadContextVar backport has correct behavior
        inside a new thread.
        """
        token = ThreadContextVar(str(uuid.uuid4()), default={'pips': 3})
        token.set({'pips': 4})

        def thread_worker():
            try:
                thread.exc_info = None

                result = token.get()

                # Test that we got a new, unmodified copy of the default
                self.assertEqual({'pips': 3}, result)

                result['pips'] = 5

                # Test that local modifications are persistent
                self.assertEqual({'pips': 5}, token.get())
            except Exception:
                import sys
                thread.exc_info = sys.exc_info()

        thread = threading.Thread(target=thread_worker)
        thread.start()
        thread.join()

        # ensure exceptions in the thread_worker fail the test
        self.assertEqual(None, thread.exc_info, thread.exc_info)

        # Test that non-local changes don't leak through
        self.assertEqual({'pips': 4}, token.get())

    def test_encoding_recursive(self):
        """
        Test that recursive data structures are replaced with '[RECURSIVE]'
        """
        data = {"Test": ["a", "b", "c"]}
        data["Self"] = data
        encoder = SanitizingJSONEncoder(logger, keyword_filters=[])
        sane_data = json.loads(encoder.encode(data))
        self.assertEqual(sane_data,
                         {"Test": ["a", "b", "c"], "Self": "[RECURSIVE]"})

    def test_encoding_recursive_repeated(self):
        """
        Test that encoding the same object twice produces the same result
        """
        data = {"Test": ["a", "b", "c"]}
        data["Self"] = data
        encoder = SanitizingJSONEncoder(logger, keyword_filters=[])
        sane_data = json.loads(encoder.encode(data))
        self.assertEqual(sane_data,
                         {"Test": ["a", "b", "c"], "Self": "[RECURSIVE]"})
        sane_data = json.loads(encoder.encode(data))
        self.assertEqual(sane_data,
                         {"Test": ["a", "b", "c"], "Self": "[RECURSIVE]"})

    def test_encoding_identical_siblings(self):
        """
        Test encoding identical siblings
        """
        inner_list = [1, 2, 3]
        nested_list = [inner_list, inner_list]
        data = {"nested0": nested_list,
                "nested1": nested_list,
                "inner": inner_list}
        encoder = SanitizingJSONEncoder(logger, keyword_filters=[])
        sane_data = json.loads(encoder.encode(data))
        self.assertEqual(sane_data,
                         {"nested0": nested_list,
                          "nested1": nested_list,
                          "inner": inner_list})

    def test_encoding_nested_repeated(self):
        """
        Test that encoding the same object within a new object is not
        incorrectly marked as recursive
        """
        encoder = SanitizingJSONEncoder(logger, keyword_filters=[])
        data = {"Test": ["a", "b", "c"]}
        encoder.encode(data)
        data = {"Previous": data, "Other": 400}
        sane_data = json.loads(encoder.encode(data))
        self.assertEqual(sane_data,
                         {"Other": 400,
                          "Previous": {"Test": ["a", "b", "c"]}})

    def test_encoding_oversized_recursive(self):
        """
        Test that encoding an object which requires trimming clips recursion
        correctly
        """
        data = {"Test": ["a" * 128 * 1024, "b", "c"], "Other": {"a": 300}}
        data["Self"] = data
        encoder = SanitizingJSONEncoder(logger, keyword_filters=[])
        sane_data = json.loads(encoder.encode(data))
        self.assertEqual(sane_data,
                         {"Test": ["a" * 1024, "b", "c"],
                          "Self": "[RECURSIVE]",
                          "Other": {"a": 300}})

    def test_encoding_time(self):
        """
        Test that encoding a large object is sufficiently speedy
        """
        setup = """\
import json
import logging
from tests.large_object import large_object_file_path
from bugsnag.utils import SanitizingJSONEncoder

logger = logging.getLogger(__name__)
encoder = SanitizingJSONEncoder(logger, keyword_filters=[])
with open(large_object_file_path()) as json_data:
    data = json.load(json_data)
        """

        stmt = "encoder.encode(data)"

        time = timeit.timeit(stmt=stmt, setup=setup, number=1000)
        maximum_time = 6

        self.assertTrue(
            time < maximum_time,
            "Encoding required {0}s (expected {1}s)".format(time, maximum_time)
        )

    def test_filter_string_values_list_handling(self):
        """
        Test that filter_string_values can accept a list for the ignored
        parameter for backwards compatibility
        """
        data = {}
        encoder = SanitizingJSONEncoder(logger)
        # no assert as we are just expecting this not to throw
        encoder.filter_string_values(data, ['password'])

    def test_sanitize_list_handling(self):
        """
        Test that _sanitize can accept a list for the ignored parameter for
        backwards compatibility
        """
        data = {}
        encoder = SanitizingJSONEncoder(logger)
        # no assert as we are just expecting this not to throw
        encoder._sanitize(data, ['password'], ['password'])

    def test_json_encode_invalid_keys(self):
        """
        Test that _sanitize can accept some invalid json where a function
        name or some other bad data is passed as a key in the payload
        dictionary.
        """
        encoder = SanitizingJSONEncoder(logger, keyword_filters=[])

        def foo():
            return "123"

        result = json.loads(encoder.encode({foo: "a"}))
        self.assertTrue(re.match(r'<function.*foo.*',
                                 list(result.keys())[0]) is not None)
        self.assertEqual(list(result.values()), ["a"])

        now = datetime.now()
        result = json.loads(encoder.encode({now: "a"}))
        self.assertEqual(list(result.keys())[0], str(now))
        self.assertEqual(list(result.values()), ["a"])

        class Object(object):
            pass

        result = json.loads(encoder.encode({Object(): "a"}))
        self.assertTrue(re.match(r'<tests.test_utils.*Object.*',
                                 list(result.keys())[0]) is not None)
        self.assertEqual(list(result.values()), ["a"])

    def test_filter_dict_with_inner_dict(self):
        """
        Test that nested dict uniqueness checks work and are not recycled
        when a reference to a nested dict goes out of scope
        """
        data = {
            'level1-key1': {
                'level2-key1': FilterDict({
                    'level3-key1': {'level4-key1': 'level4-value1'},
                    'level3-key4': {'level4-key3': 'level4-value3'},
                }),
                'level2-key2': FilterDict({
                    'level3-key2': 'level3-value1',
                    'level3-key3': {'level4-key2': 'level4-value2'},
                    'level3-key5': {'level4-key4': 'level4-value4'},
                }),
            }
        }
        encoder = SanitizingJSONEncoder(logger, keyword_filters=['password'])
        sane_data = json.loads(encoder.encode(data))
        self.assertEqual(sane_data, {
                            'level1-key1': {
                                'level2-key1': {
                                    'level3-key1': {
                                        'level4-key1': 'level4-value1'
                                    },
                                    'level3-key4': {
                                        'level4-key3': 'level4-value3'
                                    }
                                },
                                'level2-key2': {
                                    'level3-key2': 'level3-value1',
                                    'level3-key3': {
                                        'level4-key2': 'level4-value2'
                                    },
                                    'level3-key5': {
                                        'level4-key4': 'level4-value4'
                                    },
                                },
                            }
                        })

    def test_filter_strings_with_inner_dict(self):
        """
        Test that nested dict uniqueness checks work and are not recycled
        when a reference to a nested dict goes out of scope
        """
        data = FilterDict({
            'level1-key1': {
                'level2-key1': {
                    'level3-key1': {'level4-key1': 'level4-value1'},
                    'token': 'mypassword',
                },
                'level2-key2': {
                    'level3-key3': {'level4-key2': 'level4-value2'},
                    'level3-key4': {'level4-key3': 'level4-value3'},
                    'level3-key5': {'password': 'super-secret'},
                    'level3-key6': {'level4-key4': 'level4-value4'},
                    'level3-key7': {'level4-key4': 'level4-value4'},
                    'level3-key8': {'level4-key4': 'level4-value4'},
                    'level3-key9': {'level4-key4': 'level4-value4'},
                    'level3-key0': {'level4-key4': 'level4-value4'},
                },
            }
        })

        encoder = SanitizingJSONEncoder(
            logger,
            keyword_filters=['password', 'token']
        )

        filtered_data = encoder.filter_string_values(data)
        self.assertEqual(filtered_data, {
                            'level1-key1': {
                                'level2-key1': {
                                    'level3-key1': {
                                        'level4-key1': 'level4-value1'
                                    },
                                    'token': '[FILTERED]'
                                },
                                'level2-key2': {
                                    'level3-key3': {
                                        'level4-key2': 'level4-value2'
                                    },
                                    'level3-key4': {
                                        'level4-key3': 'level4-value3'
                                    },
                                    'level3-key5': {
                                        'password': '[FILTERED]'
                                    },
                                    'level3-key6': {
                                        'level4-key4': 'level4-value4'
                                    },
                                    'level3-key7': {
                                        'level4-key4': 'level4-value4'
                                    },
                                    'level3-key8': {
                                        'level4-key4': 'level4-value4'
                                    },
                                    'level3-key9': {
                                        'level4-key4': 'level4-value4'
                                    },
                                    'level3-key0': {
                                        'level4-key4': 'level4-value4'
                                    },
                                },
                            }
                        })

    def test_parse_invalid_content_type(self):
        info = parse_content_type('invalid-type')
        self.assertEqual(('invalid-type', None, None, None), info)

    def test_parse_invalid_content_type_params(self):
        info = parse_content_type('invalid-type;schema=http://example.com/b')
        self.assertEqual(('invalid-type', None, None,
                          'schema=http://example.com/b'), info)

    def test_parse_parameters(self):
        info = parse_content_type('text/plain;charset=utf-32')
        self.assertEqual(('text', 'plain', None, 'charset=utf-32'), info)

    def test_parse_suffix(self):
        info = parse_content_type('application/hal+json;charset=utf-8')
        self.assertEqual(('application', 'hal', 'json', 'charset=utf-8'), info)

    def test_json_content_type(self):
        self.assertTrue(is_json_content_type('application/json'))
        self.assertTrue(is_json_content_type('application/hal+json'))
        self.assertTrue(is_json_content_type('application/other+json'))
        self.assertTrue(is_json_content_type(
            'application/schema+json;schema=http://example.com/schema-2'))
        self.assertTrue(is_json_content_type('application/json;charset=utf-8'))
        self.assertFalse(is_json_content_type('text/json'))
        self.assertFalse(is_json_content_type('text/plain'))
        self.assertFalse(is_json_content_type('json'))
        self.assertFalse(is_json_content_type('application/jsonfoo'))


utc = timezone.utc
plus_1 = timezone(timedelta(hours=1))
plus_12 = timezone(timedelta(hours=12))
plus_0_15 = timezone(timedelta(hours=0, minutes=15))
plus_12_30 = timezone(timedelta(hours=12, minutes=30))
minus_1 = timezone(timedelta(hours=-1))
minus_12 = timezone(timedelta(hours=-12))
minus_0_15 = timezone(timedelta(hours=0, minutes=-15))
minus_12_30 = timezone(timedelta(hours=-12, minutes=-30))


@pytest.mark.parametrize("dt, expected", [
    (datetime(1, 1, 1, 1, 1, 1, 1), '0001-01-01T01:01:01.000'),  # noqa: E501
    (datetime(1900, 1, 2, 3, 4, 5, 678900), '1900-01-02T03:04:05.678'),  # noqa: E501
    (datetime(9999, 12, 31, 23, 59, 59, 999999), '9999-12-31T23:59:59.999'),  # noqa: E501
    (datetime(1950, 5, 14, 20, 34, 52, 61000), '1950-05-14T20:34:52.061'),  # noqa: E501
    (datetime(1999, 4, 3, 14, 7, 12), '1999-04-03T14:07:12.000'),  # noqa: E501
    (datetime(2021, 1, 1, tzinfo=timezone.utc), '2021-01-01T00:00:00.000+00:00'),  # noqa: E501
    (datetime(1, 1, 1, 1, 1, 1, 1, tzinfo=plus_1), '0001-01-01T01:01:01.000+01:00'),  # noqa: E501
    (datetime(1900, 1, 2, 3, 4, 5, 678900, tzinfo=plus_12), '1900-01-02T03:04:05.678+12:00'),  # noqa: E501
    (datetime(1999, 4, 3, 14, 7, 12, tzinfo=plus_0_15), '1999-04-03T14:07:12.000+00:15'),  # noqa: E501
    (datetime(1950, 5, 14, 20, 34, 52, 61000, tzinfo=plus_12_30), '1950-05-14T20:34:52.061+12:30'),  # noqa: E501
    (datetime(9999, 12, 31, 23, 59, 59, 999999, tzinfo=minus_1), '9999-12-31T23:59:59.999-01:00'),  # noqa: E501
    (datetime(1950, 5, 14, 20, 34, 52, 61000, tzinfo=minus_12), '1950-05-14T20:34:52.061-12:00'),  # noqa: E501
    (datetime(1999, 4, 3, 14, 7, 12, tzinfo=minus_0_15), '1999-04-03T14:07:12.000-00:15'),  # noqa: E501
    (datetime(1950, 5, 14, 20, 34, 52, 61000, tzinfo=minus_12_30), '1950-05-14T20:34:52.061-12:30'),  # noqa: E501
])
def test_to_rfc3339(dt: datetime, expected: str):
    assert to_rfc3339(dt) == expected


@pytest.mark.parametrize("url, expected", [
    ('https://example.com', 'https://example.com'),
    ('https://example.com/', 'https://example.com/'),
    ('https://example.com/a/b/c', 'https://example.com/a/b/c'),
    ('https://example.com/abc?xyz=123', 'https://example.com/abc'),
    ('https://example.com/a/b/c', 'https://example.com/a/b/c'),
    ('https://example.com/abc?xyz=123', 'https://example.com/abc'),
    ('https://example.com/abc;x=1;y=2;z=3', 'https://example.com/abc'),
    ('https://example.com:8000/abc?xyz=123', 'https://example.com:8000/abc'),
    (b'https://example.com', b'https://example.com'),
    (b'https://example.com/', b'https://example.com/'),
    (b'https://example.com/a/b/c', b'https://example.com/a/b/c'),
    (b'https://example.com/abc?xyz=123', b'https://example.com/abc'),
    (b'https://example.com/a/b/c', b'https://example.com/a/b/c'),
    (b'https://example.com/abc?xyz=123', b'https://example.com/abc'),
    (b'https://example.com/abc;x=1;y=2;z=3', b'https://example.com/abc'),
    (b'https://example.com:8000/abc?xyz=123', b'https://example.com:8000/abc'),

    ('wss://example.com/abc?xyz=123', 'wss://example.com/abc'),
    ('ftp://example.com/abc?xyz=123', 'ftp://example.com/abc'),
    (b'wss://example.com/abc?xyz=123', b'wss://example.com/abc'),
    (b'ftp://example.com/abc?xyz=123', b'ftp://example.com/abc'),

    ('xyz', 'xyz'),
    ('///', '/'),
    ('/a/b/c', '/a/b/c'),
    ('example.com/<<<<', 'example.com/<<<<'),
    ('->example.com<-', '->example.com<-'),
    (b'xyz', b'xyz'),
    (b'///', b'/'),
    (b'/a/b/c', b'/a/b/c'),
    (b'example.com/<<<<', b'example.com/<<<<'),
    (b'->example.com<-', b'->example.com<-'),

    ('', None),
    (b'', None),
    ('  ', None),
    (b'  ', None),
    (None, None),
    (123, None),
    ([1, 2, 3], None),
    ({'a': 1, b'b': 2, 'c': 3}, None),
    (object(), None),
    (lambda: 'example.com', None),
])
def test_remove_query_from_url(url, expected):
    assert remove_query_from_url(url) == expected



import bugsnag
import pytest
import sys

from tests import fixtures
from .utils import is_exception_group_supported


# the exceptiongroup package will only define ExceptionGroup on versions <3.11
# on 3.11+ it re-exports the built-in class so even if exceptiongroup is
# installed, the class name won't have the 'exceptiongroup' prefix
if sys.version_info < (3, 11):
    expected_exception_group_class = 'exceptiongroup.ExceptionGroup'
else:
    expected_exception_group_class = 'ExceptionGroup'


@pytest.mark.skipif(
    not is_exception_group_supported,
    reason="requires python 3.11 or the exceptiongroup pacakge"
)
def test_exception_groups_are_unwrapped(bugsnag_server):
    # disable send_code so we can assert against stacktraces more easily
    bugsnag.configure(send_code=False)

    bugsnag.notify(fixtures.exception_group_with_no_cause)

    bugsnag_server.wait_for_request()
    assert bugsnag_server.sent_report_count == 1

    payload = bugsnag_server.received[0]['json_body']
    exceptions = payload['events'][0]['exceptions']

    # there should be 1 ExceptionGroup with 4 sub-exceptions
    assert len(exceptions) == 5

    assert exceptions[0] == {
        'message': 'the message of the group (4 sub-exceptions)',
        'errorClass': expected_exception_group_class,
        'type': 'python',
        'stacktrace': [
            {
                'file': 'tests/fixtures/exception_groups.py',
                'method': 'raise_exception_group_with_no_cause',
                'lineNumber': 21,
                'inProject': True,
                'code': None,
            },
            {
                'file': 'tests/fixtures/exception_groups.py',
                'method': '<module>',
                'lineNumber': 25,
                'inProject': True,
                'code': None,
            },
        ],
    }

    expected_sub_exceptions = [
        'Exception',
        'ArithmeticError',
        'NameError',
        'AssertionError',
    ]

    for i, expected in enumerate(expected_sub_exceptions, start=1):
        assert exceptions[i] == {
            'message': 'exception #' + str(i),
            'errorClass': expected,
            'type': 'python',
            'stacktrace': [
                {
                    'file': 'tests/fixtures/exception_groups.py',
                    'method': 'generate_exception',
                    'lineNumber': 15,
                    'inProject': True,
                    'code': None,
                }
            ]
        }


@pytest.mark.skipif(
    not is_exception_group_supported,
    reason="requires python 3.11 or the exceptiongroup pacakge"
)
def test_base_exception_group_subclasses_are_unwrapped(bugsnag_server):
    # disable send_code so we can assert against stacktraces more easily
    bugsnag.configure(send_code=False)

    bugsnag.notify(fixtures.base_exception_group_subclass)

    bugsnag_server.wait_for_request()
    assert bugsnag_server.sent_report_count == 1

    payload = bugsnag_server.received[0]['json_body']
    exceptions = payload['events'][0]['exceptions']

    # there should be 1 MyExceptionGroup with 3 sub-exceptions
    assert len(exceptions) == 4

    assert exceptions[0] == {
        'message': 'my very easy method just speeds up (n)making exception groups (3 sub-exceptions)',  # noqa: E501
        'errorClass': 'tests.fixtures.exception_groups.MyExceptionGroup',
        'type': 'python',
        'stacktrace': [
            {
                'file': 'tests/fixtures/exception_groups.py',
                'method': 'raise_base_exception_group_subclass_with_no_cause',  # noqa: E501
                'lineNumber': 35,
                'inProject': True,
                'code': None,
            },
            {
                'file': 'tests/fixtures/exception_groups.py',
                'method': '<module>',
                'lineNumber': 39,
                'inProject': True,
                'code': None,
            },
        ],
    }

    expected_sub_exceptions = [
        'GeneratorExit',
        'ReferenceError',
        'NotImplementedError',
    ]

    for i, expected in enumerate(expected_sub_exceptions, start=1):
        assert exceptions[i] == {
            'message': 'exception #' + str(i),
            'errorClass': expected,
            'type': 'python',
            'stacktrace': [
                {
                    'file': 'tests/fixtures/exception_groups.py',
                    'method': 'generate_exception',
                    'lineNumber': 15,
                    'inProject': True,
                    'code': None,
                }
            ]
        }


@pytest.mark.skipif(
    not is_exception_group_supported,
    reason="requires python 3.11 or the exceptiongroup pacakge"
)
def test_do_not_recurse_into_sub_exception_groups(bugsnag_server):
    # disable send_code so we can assert against stacktraces more easily
    bugsnag.configure(send_code=False)

    bugsnag.notify(fixtures.exception_group_with_nested_group)

    bugsnag_server.wait_for_request()
    assert bugsnag_server.sent_report_count == 1

    payload = bugsnag_server.received[0]['json_body']
    exceptions = payload['events'][0]['exceptions']

    # there should be 1 ExceptionGroup with 3 sub-exceptions
    assert len(exceptions) == 4

    assert exceptions[0] == {
        'message': 'the message of the group (3 sub-exceptions)',
        'errorClass': expected_exception_group_class,
        'type': 'python',
        'stacktrace': [
            {
                'file': 'tests/fixtures/exception_groups.py',
                'method': 'raise_exception_group_with_nested_group',
                'lineNumber': 45,
                'inProject': True,
                'code': None,
            },
            {
                'file': 'tests/fixtures/exception_groups.py',
                'method': '<module>',
                'lineNumber': 49,
                'inProject': True,
                'code': None,
            },
        ],
    }

    assert exceptions[1] == {
        'message': 'exception #1',
        'errorClass': 'Exception',
        'type': 'python',
        'stacktrace': [
            {
                'file': 'tests/fixtures/exception_groups.py',
                'method': 'generate_exception',
                'lineNumber': 15,
                'inProject': True,
                'code': None,
            }
        ]
    }

    assert exceptions[2] == {
        'message': 'the message of the group (4 sub-exceptions)',
        'errorClass': expected_exception_group_class,
        'type': 'python',
        'stacktrace': [
            {
                'file': 'tests/fixtures/exception_groups.py',
                'method': 'raise_exception_group_with_no_cause',
                'lineNumber': 21,
                'inProject': True,
                'code': None,
            },
            {
                'file': 'tests/fixtures/exception_groups.py',
                'method': '<module>',
                'lineNumber': 25,
                'inProject': True,
                'code': None,
            },
        ],
    }

    assert exceptions[3] == {
        'message': 'exception #3',
        'errorClass': 'ArithmeticError',
        'type': 'python',
        'stacktrace': [
            {
                'file': 'tests/fixtures/exception_groups.py',
                'method': 'generate_exception',
                'lineNumber': 15,
                'inProject': True,
                'code': None,
            }
        ]
    }


@pytest.mark.skipif(
    not is_exception_group_supported,
    reason="requires python 3.11 or the exceptiongroup pacakge"
)
def test_exception_group_implicit_cause_is_traversed(bugsnag_server):
    # disable send_code so we can assert against stacktraces more easily
    bugsnag.configure(send_code=False)

    bugsnag.notify(fixtures.exception_group_with_implicit_cause)

    bugsnag_server.wait_for_request()
    assert bugsnag_server.sent_report_count == 1

    payload = bugsnag_server.received[0]['json_body']
    exceptions = payload['events'][0]['exceptions']

    print(exceptions)

    # there should be 1 ExceptionGroup with 1 cause and 2 sub-exceptions
    assert len(exceptions) == 4

    # the ExceptionGroup
    assert exceptions[0] == {
        'message': 'group with implicit cause (2 sub-exceptions)',
        'errorClass': expected_exception_group_class,
        'type': 'python',
        'stacktrace': [
            {
                'file': 'tests/fixtures/exception_groups.py',
                'method': 'raise_exception_group_with_implicit_cause',
                'lineNumber': 58,
                'inProject': True,
                'code': None,
            },
            {
                'file': 'tests/fixtures/exception_groups.py',
                'method': '<module>',
                'lineNumber': 62,
                'inProject': True,
                'code': None,
            },
        ],
    }

    # the cause - another ExceptionGroup
    # note: we don't recurse into this!
    assert exceptions[1] == {
        'message': 'the message of the group (3 sub-exceptions)',
        'errorClass': expected_exception_group_class,
        'type': 'python',
        'stacktrace': [
            {
                'file': 'tests/fixtures/exception_groups.py',
                'method': 'raise_exception_group_with_nested_group',
                'lineNumber': 45,
                'inProject': True,
                'code': None,
            },
            {
                'file': 'tests/fixtures/exception_groups.py',
                'method': 'raise_exception_group_with_implicit_cause',
                'lineNumber': 56,
                'inProject': True,
                'code': None,
            },
        ],
    }

    # the exceptions in the original ExceptionGroup
    assert exceptions[2] == {
        'message': 'a',
        'errorClass': 'NameError',
        'type': 'python',
        'stacktrace': [
            {
                'file': 'tests/fixtures/caused_by.py',
                'method': 'raise_exception_with_explicit_cause',
                'lineNumber': 5,
                'inProject': True,
                'code': None,
            },
            {
                'file': 'tests/fixtures/caused_by.py',
                'method': '<module>',
                'lineNumber': 20,
                'inProject': True,
                'code': None,
            },
        ],
    }

    assert exceptions[3] == {
        'message': 'exception #2',
        'errorClass': 'NameError',
        'type': 'python',
        'stacktrace': [
            {
                'file': 'tests/fixtures/exception_groups.py',
                'method': 'generate_exception',
                'lineNumber': 15,
                'inProject': True,
                'code': None,
            },
        ],
    }


@pytest.mark.skipif(
    not is_exception_group_supported,
    reason="requires python 3.11 or the exceptiongroup pacakge"
)
def test_exception_group_explicit_cause_is_traversed(bugsnag_server):
    # disable send_code so we can assert against stacktraces more easily
    bugsnag.configure(send_code=False)

    bugsnag.notify(fixtures.exception_group_with_explicit_cause)

    bugsnag_server.wait_for_request()
    assert bugsnag_server.sent_report_count == 1

    payload = bugsnag_server.received[0]['json_body']
    exceptions = payload['events'][0]['exceptions']

    print(exceptions)

    # 1 ExceptionGroup + 1 cause + 1 sub-cause + 2 sub-exceptions
    assert len(exceptions) == 5

    assert exceptions[0] == {
        'message': 'group with explicit cause (2 sub-exceptions)',
        'errorClass': expected_exception_group_class,
        'type': 'python',
        'stacktrace': [
            {
                'file': 'tests/fixtures/exception_groups.py',
                'method': 'raise_exception_group_with_explicit_cause',
                'lineNumber': 71,
                'inProject': True,
                'code': None,
            },
            {
                'file': 'tests/fixtures/exception_groups.py',
                'method': '<module>',
                'lineNumber': 75,
                'inProject': True,
                'code': None,
            },
        ],
    }

    assert exceptions[1] == {
        'message': 'group with implicit cause (2 sub-exceptions)',
        'errorClass': expected_exception_group_class,
        'type': 'python',
        'stacktrace': [
            {
                'file': 'tests/fixtures/exception_groups.py',
                'method': 'raise_exception_group_with_implicit_cause',
                'lineNumber': 58,
                'inProject': True,
                'code': None,
            },
            {
                'file': 'tests/fixtures/exception_groups.py',
                'method': 'raise_exception_group_with_explicit_cause',
                'lineNumber': 69,
                'inProject': True,
                'code': None,
            },
        ],
    }

    assert exceptions[2] == {
        'message': 'the message of the group (3 sub-exceptions)',
        'errorClass': expected_exception_group_class,
        'type': 'python',
        'stacktrace': [
            {
                'file': 'tests/fixtures/exception_groups.py',
                'method': 'raise_exception_group_with_nested_group',
                'lineNumber': 45,
                'inProject': True,
                'code': None,
            },
            {
                'file': 'tests/fixtures/exception_groups.py',
                'method': 'raise_exception_group_with_implicit_cause',
                'lineNumber': 56,
                'inProject': True,
                'code': None,
            },
        ],
    }

    assert exceptions[3] == {
        'message': 'exception #1',
        'errorClass': 'NameError',
        'type': 'python',
        'stacktrace': [
            {
                'file': 'tests/fixtures/exception_groups.py',
                'method': 'generate_exception',
                'lineNumber': 15,
                'inProject': True,
                'code': None,
            },
        ],
    }

    assert exceptions[4] == {
        'message': 'a',
        'errorClass': 'NameError',
        'type': 'python',
        'stacktrace': [
            {
                'file': 'tests/fixtures/caused_by.py',
                'method': 'raise_exception_with_explicit_cause',
                'lineNumber': 5,
                'inProject': True,
                'code': None,
            },
            {
                'file': 'tests/fixtures/caused_by.py',
                'method': '<module>',
                'lineNumber': 20,
                'inProject': True,
                'code': None,
            },
        ],
    }



import pytest

from bugsnag import RequestConfiguration


def test_request_config_meta_data_warns():
    config = RequestConfiguration.get_instance()

    with pytest.warns(DeprecationWarning) as records:
        config.meta_data['foo'] = 'bar'
        assert len(records) == 1
        assert str(records[0].message) == ('RequestConfiguration.meta_data ' +
                                           'has been renamed to "metadata"')
    assert config.metadata['foo'] == 'bar'



import sys
import pytest
import asyncio
from time import sleep
from threading import Thread
from random import shuffle, randrange
from bugsnag.client import Client
from bugsnag.feature_flags import FeatureFlag
from bugsnag.context import create_new_context


def test_state_is_stored_separately_per_thread():
    client1 = Client()
    client2 = Client()

    assert client1.feature_flags == []
    assert client2.feature_flags == []

    client1.add_feature_flag('a')
    client2.add_feature_flag('b')

    assert client1.feature_flags == [FeatureFlag('a')]
    assert client2.feature_flags == [FeatureFlag('b')]

    def thread_target(client1, client2, id):
        thread.exception = None

        try:
            # sleep for a bit to allow other threads time to interfere
            sleep(randrange(0, 100) / 1000)

            assert client1.feature_flags == []
            assert client2.feature_flags == []

            client1.add_feature_flag('client1 (#%d) 1' % id)
            client2.add_feature_flag('client2 (#%d) 1' % id)

            # sleep for a bit to allow other threads time to interfere
            sleep(randrange(0, 100) / 1000)

            client1.add_feature_flag('client1 (#%d) 2' % id)
            client2.add_feature_flag('client2 (#%d) 2' % id)

            assert client1.feature_flags == [
                FeatureFlag('client1 (#%d) 1' % id),
                FeatureFlag('client1 (#%d) 2' % id)
            ]

            assert client2.feature_flags == [
                FeatureFlag('client2 (#%d) 1' % id),
                FeatureFlag('client2 (#%d) 2' % id)
            ]

        except Exception as e:
            thread.exception = e

    threads = []

    for i in range(10):
        thread = Thread(target=thread_target, args=(client1, client2, i))
        threads.append(thread)

    shuffle(threads)

    for thread in threads:
        thread.start()

    shuffle(threads)

    for thread in threads:
        thread.join(2)

        assert not thread.is_alive()

        if thread.exception is not None:
            raise thread.exception

    # changes in other threads should not affect this thread
    assert client1.feature_flags == [FeatureFlag('a')]
    assert client2.feature_flags == [FeatureFlag('b')]


@pytest.mark.skipif(
    sys.version_info < (3, 7),
    reason="requires ContextVar support (Python 3.7 or higher)"
)
def test_state_is_stored_separately_per_async_context():
    client1 = Client()
    client2 = Client()

    assert client1.feature_flags == []
    assert client2.feature_flags == []

    client1.add_feature_flag('a')
    client2.add_feature_flag('b')

    assert client1.feature_flags == [FeatureFlag('a')]
    assert client2.feature_flags == [FeatureFlag('b')]

    async def mutate_state(id):
        create_new_context()

        await asyncio.sleep(randrange(0, 100) / 1000)

        assert client1.feature_flags == []
        assert client2.feature_flags == []

        client1.add_feature_flag('client1 (#%d) 1' % id)
        await asyncio.sleep(randrange(0, 100) / 1000)

        client2.add_feature_flag('client2 (#%d) 1' % id)
        await asyncio.sleep(randrange(0, 100) / 1000)

        assert client1.feature_flags == [FeatureFlag('client1 (#%d) 1' % id)]
        assert client2.feature_flags == [FeatureFlag('client2 (#%d) 1' % id)]

        client1.add_feature_flag('client1 (#%d) 2' % id)
        client2.add_feature_flag('client2 (#%d) 2' % id)

        await asyncio.sleep(randrange(0, 100) / 1000)

        assert client1.feature_flags == [
            FeatureFlag('client1 (#%d) 1' % id),
            FeatureFlag('client1 (#%d) 2' % id),
        ]

        assert client2.feature_flags == [
            FeatureFlag('client2 (#%d) 1' % id),
            FeatureFlag('client2 (#%d) 2' % id),
        ]

    async def test():
        tasks = [mutate_state(i) for i in range(10)]

        await asyncio.gather(*tasks)

    loop = asyncio.new_event_loop()

    try:
        loop.run_until_complete(test())
    finally:
        loop.close()

    assert client1.feature_flags == [FeatureFlag('a')]
    assert client2.feature_flags == [FeatureFlag('b')]



import unittest
import threading

from bugsnag.sessiontracker import SessionTracker, SessionMiddleware
from bugsnag.configuration import Configuration
from bugsnag.event import Event


class TestSessionMiddleware(unittest.TestCase):
    def setUp(self):
        self.config = Configuration()
        self.config.configure(api_key='fff', auto_capture_sessions=False)
        self.sessiontracker = SessionTracker(self.config)
        self.sessiontracker.auto_sessions = True  # Stub session delivery queue

    def tearDown(self):
        pass

    def test_increment_counts(self):
        """
        Every event should keep a list of prior events which occurred in the
        session
        """

        def next_callable(event):
            pass

        middleware = SessionMiddleware(next_callable)
        self.sessiontracker.start_session()

        event = Event(Exception('shucks'), self.config, None)
        middleware(event)

        assert event.session['events']['unhandled'] == 0
        assert event.session['events']['handled'] == 1

        event2 = Event(Exception('oh no'), self.config, None)
        middleware(event2)

        assert event2.session['events']['unhandled'] == 0
        assert event2.session['events']['handled'] == 2

        # Session counts should not change for events already handled
        assert event.session['events']['unhandled'] == 0
        assert event.session['events']['handled'] == 1

    def test_it_does_nothing_if_no_session_exists(self):
        def run_session_middleware():
            def next_callable(event):
                pass

            try:
                thread.exc_info = None
                middleware = SessionMiddleware(next_callable)

                event = Event(Exception('shucks'), self.config, None)
                middleware(event)

                assert event.session is None
            except Exception:
                import sys
                thread.exc_info = sys.exc_info()

        thread = threading.Thread(target=run_session_middleware)
        thread.start()
        thread.join()

        # ensure there was no exception in the thread
        self.assertEqual(None, thread.exc_info, thread.exc_info)



import platform
import time

from bugsnag import Client
from bugsnag.configuration import Configuration
from bugsnag.notifier import _NOTIFIER_INFORMATION
from bugsnag.sessiontracker import SessionTracker
from tests.utils import IntegrationTest


class TestConfiguration(IntegrationTest):
    def setUp(self):
        super(TestConfiguration, self).setUp()
        self.config = Configuration()
        self.config.auto_capture_sessions = True

    def test_session_tracker_adds_session_object_to_queue(self):
        tracker = SessionTracker(self.config)
        tracker.auto_sessions = True
        tracker.start_session()
        self.assertEqual(len(tracker.session_counts), 1)
        for key, value in tracker.session_counts.items():
            self.assertEqual(value, 1)

    def test_session_tracker_send_sessions_sends_sessions(self):
        client = Client(
            auto_capture_sessions=True,
            session_endpoint=self.server.url,
            asynchronous=False
        )
        client.session_tracker.start_session()
        self.assertEqual(len(client.session_tracker.session_counts), 1)
        client.session_tracker.send_sessions()
        self.assertEqual(len(client.session_tracker.session_counts), 0)
        json_body = self.server.received[0]['json_body']
        self.assertTrue('app' in json_body)
        self.assertTrue('notifier' in json_body)
        self.assertTrue('device' in json_body)
        self.assertTrue('sessionCounts' in json_body)
        self.assertEqual(len(json_body['sessionCounts']), 1)

    def test_session_tracker_sets_details_from_config(self):
        client = Client(
            auto_capture_sessions=True,
            session_endpoint=self.server.url,
            asynchronous=False
        )
        client.session_tracker.start_session()
        client.session_tracker.send_sessions()
        json_body = self.server.received[0]['json_body']

        # Notifier properties
        assert json_body['notifier'] == _NOTIFIER_INFORMATION

        # App properties
        app = json_body['app']
        self.assertTrue('releaseStage' in app)
        self.assertEqual(app['releaseStage'],
                         client.configuration.release_stage)
        self.assertTrue('version' in app)
        self.assertEqual(app['version'],
                         client.configuration.app_version)
        # Device properties
        device = json_body['device']
        self.assertTrue('hostname' in device)
        self.assertEqual(device['hostname'],
                         client.configuration.hostname)
        self.assertTrue('runtimeVersions' in device)
        self.assertEqual(device['runtimeVersions']['python'],
                         platform.python_version())

    def test_session_middleware_attaches_session_to_event(self):
        client = Client(
            auto_capture_sessions=True,
            session_endpoint=self.server.url + '/ignore',
            endpoint=self.server.url,
            asynchronous=False
        )
        client.session_tracker.start_session()
        client.notify(Exception("Test"))
        while len(self.server.received) == 0:
            time.sleep(0.5)
        json_body = self.server.received[0]['json_body']
        event = json_body['events'][0]
        self.assertTrue('session' in event)
        session = event['session']
        self.assertTrue('id' in session)
        self.assertTrue('startedAt' in session)
        self.assertTrue('events' in session)
        sesevents = session['events']
        self.assertTrue('unhandled' in sesevents)
        self.assertEqual(sesevents['unhandled'], 0)
        self.assertTrue('handled' in sesevents)
        self.assertEqual(sesevents['handled'], 1)

    def test_session_tracker_does_not_send_when_nothing_to_send(self):
        client = Client(
            api_key='fff',
            session_endpoint=self.server.url,
            asynchronous=False,
            release_stage="dev",
            notify_release_stages=["prod"],
        )
        client.session_tracker.send_sessions()
        self.assertEqual(0, len(self.server.received))

    def test_session_tracker_does_not_send_when_disabled(self):
        client = Client(
            api_key='fff',
            session_endpoint=self.server.url,
            asynchronous=False,
            release_stage="dev",
            notify_release_stages=["prod"],
        )
        client.session_tracker.start_session()
        client.session_tracker.send_sessions()
        self.assertEqual(0, len(self.server.received))

    def test_session_tracker_does_not_send_when_misconfigured(self):
        client = Client(
            api_key=None,
            session_endpoint=self.server.url,
            asynchronous=False,
            release_stage="dev",
            notify_release_stages=["prod"],
        )
        client.session_tracker.start_session()
        client.session_tracker.send_sessions()
        self.assertEqual(0, len(self.server.received))



from bugsnag.feature_flags import FeatureFlag, FeatureFlagDelegate
import pytest


class Unstringable:
    def __str__(self):
        raise Exception('no')

    def __repr__(self):
        raise Exception('nope')


_invalid_names = [
    (None,),
    (True,),
    (False,),
    (1234,),
    ([1, 2, 3],),
    ({'a': 1, 'b': 2},),
    ('',)
]


def test_feature_flag_has_name_and_variant():
    flag = FeatureFlag('abc', 'xyz')

    assert flag.name == 'abc'
    assert flag.variant == 'xyz'


def test_feature_flag_variant_is_optional():
    assert FeatureFlag('a').variant is None
    assert FeatureFlag('a', None).variant is None


def test_feature_flag_variant_is_coerced_to_string():
    assert FeatureFlag('a', 123).variant == '123'
    assert FeatureFlag('a', [1, 2, 3]).variant == '[1, 2, 3]'


def test_feature_flag_name_and_variant_can_be_bytes():
    flag = FeatureFlag(b'abc', b'xyz')

    assert flag.name == b'abc'
    assert flag.variant == b'xyz'


def test_feature_flag_variant_is_unset_if_not_coercable():
    assert FeatureFlag('a', Unstringable()).variant is None


def test_feature_flag_implements_repr():
    assert repr(FeatureFlag('abc', 'xyz')) == "FeatureFlag('abc', 'xyz')"
    assert repr(FeatureFlag('ayc', None)) == "FeatureFlag('ayc', None)"
    assert repr(FeatureFlag('xyz')) == "FeatureFlag('xyz', None)"


def test_feature_flag_can_be_converted_to_dict():
    flag = FeatureFlag('abc', 'xyz')

    assert flag.to_dict() == {'featureFlag': 'abc', 'variant': 'xyz'}


def test_feature_flag_dict_does_not_have_variant_when_variant_is_not_given():
    flag = FeatureFlag('xyz')

    assert flag.to_dict() == {'featureFlag': 'xyz'}


def test_feature_flag_dict_does_not_have_variant_when_variant_is_none():
    flag = FeatureFlag('abc', variant=None)

    assert flag.to_dict() == {'featureFlag': 'abc'}


def test_a_feature_flag_with_name_and_variant_is_valid():
    assert FeatureFlag('abc', 'xyz').is_valid() is True
    assert FeatureFlag('abc', b'xyz').is_valid() is True
    assert FeatureFlag(b'abc', b'xyz').is_valid() is True
    assert FeatureFlag(b'abc', 'xyz').is_valid() is True


def test_a_feature_flag_with_only_name_is_valid():
    flag = FeatureFlag('b')

    assert flag.is_valid() is True


def test_a_feature_flag_with_empty_name_is_not_valid():
    flag = FeatureFlag('')

    assert flag.is_valid() is False


def test_a_feature_flag_with_none_name_is_not_valid():
    flag = FeatureFlag(None)

    assert flag.is_valid() is False


def test_feature_flags_are_comparable():
    flag1 = FeatureFlag('a')
    flag2 = FeatureFlag('b', 'x')
    flag3 = FeatureFlag('c')

    assert flag1 == FeatureFlag('a')
    assert flag2 == FeatureFlag('b', 'x')
    assert flag3 == FeatureFlag('c')

    assert flag1 != FeatureFlag('a', '1')
    assert flag2 != FeatureFlag('b')

    assert flag1 != flag2
    assert flag1 != flag3
    assert flag2 != flag3

    assert flag1 != 'a'
    assert flag1 is not None


def test_delegate_contains_no_flags_by_default():
    delegate = FeatureFlagDelegate()

    assert delegate.to_list() == []
    assert delegate.to_json() == []


def test_delegate_does_not_get_mutated_after_being_copied():
    delegate1 = FeatureFlagDelegate()
    delegate1.add('abc', '123')

    delegate2 = delegate1.copy()
    delegate2.add('xyz', '456')

    delegate3 = delegate2.copy()
    delegate3.clear()

    assert delegate1.to_json() == [{'featureFlag': 'abc', 'variant': '123'}]
    assert delegate2.to_json() == [
        {'featureFlag': 'abc', 'variant': '123'},
        {'featureFlag': 'xyz', 'variant': '456'},
    ]
    assert delegate3.to_json() == []


def test_delegate_can_add_flags_individually():
    delegate = FeatureFlagDelegate()
    delegate.add('abc', 'xyz')
    delegate.add('another', None)
    delegate.add('again', 123)

    assert delegate.to_json() == [
        {'featureFlag': 'abc', 'variant': 'xyz'},
        {'featureFlag': 'another'},
        {'featureFlag': 'again', 'variant': '123'}
    ]


def test_delegate_add_replaces_by_name_when_the_original_has_no_variant():
    delegate = FeatureFlagDelegate()
    delegate.add('abc', None)
    delegate.add('another', None)
    delegate.add('abc', 123)

    assert delegate.to_json() == [
        {'featureFlag': 'abc', 'variant': '123'},
        {'featureFlag': 'another'}
    ]


def test_delegate_add_replaces_by_name_when_the_replacement_has_no_variant():
    delegate = FeatureFlagDelegate()
    delegate.add('abc', '123')
    delegate.add('another', None)
    delegate.add('abc', None)

    assert delegate.to_json() == [
        {'featureFlag': 'abc'},
        {'featureFlag': 'another'}
    ]


def test_delegate_add_replaces_by_name_when_both_have_variants():
    delegate = FeatureFlagDelegate()
    delegate.add('abc', '123')
    delegate.add('another', None)
    delegate.add('abc', '456')

    assert delegate.to_json() == [
        {'featureFlag': 'abc', 'variant': '456'},
        {'featureFlag': 'another'}
    ]


@pytest.mark.parametrize('invalid_name', _invalid_names)
def test_delegate_add_add_drops_flag_when_name_is_invalid(invalid_name):
    delegate = FeatureFlagDelegate()
    delegate.add('abc', '123')
    delegate.add(invalid_name, None)

    assert delegate.to_json() == [
        {'featureFlag': 'abc', 'variant': '123'}
    ]


def test_delegate_can_add_multiple_flags_at_once():
    delegate = FeatureFlagDelegate()
    delegate.merge([
        FeatureFlag('a', 'xyz'),
        FeatureFlag('b'),
        FeatureFlag('c', '111'),
        FeatureFlag('d'),
    ])

    assert delegate.to_json() == [
        {'featureFlag': 'a', 'variant': 'xyz'},
        {'featureFlag': 'b'},
        {'featureFlag': 'c', 'variant': '111'},
        {'featureFlag': 'd'}
    ]


def test_delegate_can_merge_new_flags_with_existing_ones():
    delegate = FeatureFlagDelegate()
    delegate.merge([
        FeatureFlag('a', 'xyz'),
        FeatureFlag('b'),
        FeatureFlag('c', '111'),
        FeatureFlag('d'),
    ])

    delegate.merge([
        FeatureFlag('e'),
        FeatureFlag('a'),
        FeatureFlag('d', 'xyz'),
    ])

    assert delegate.to_json() == [
        {'featureFlag': 'a'},
        {'featureFlag': 'b'},
        {'featureFlag': 'c', 'variant': '111'},
        {'featureFlag': 'd', 'variant': 'xyz'},
        {'featureFlag': 'e'}
    ]


def test_delegate_merge_replaces_by_name_when_the_original_has_no_variant():
    delegate = FeatureFlagDelegate()

    delegate.add('a', None)
    delegate.merge([
        FeatureFlag('a', 'xyz'),
        FeatureFlag('b'),
        FeatureFlag('c', '111'),
        FeatureFlag('d'),
    ])

    assert delegate.to_json() == [
        {'featureFlag': 'a', 'variant': 'xyz'},
        {'featureFlag': 'b'},
        {'featureFlag': 'c', 'variant': '111'},
        {'featureFlag': 'd'}
    ]


def test_delegate_merge_replaces_by_name_when_the_replacement_has_no_variant():
    delegate = FeatureFlagDelegate()

    delegate.add('a', 'xyz')
    delegate.merge([
        FeatureFlag('a', None),
        FeatureFlag('b'),
        FeatureFlag('c', '111'),
        FeatureFlag('d'),
    ])

    assert delegate.to_json() == [
        {'featureFlag': 'a'},
        {'featureFlag': 'b'},
        {'featureFlag': 'c', 'variant': '111'},
        {'featureFlag': 'd'}
    ]


def test_delegate_merge_replaces_by_name_when_both_have_variants():
    delegate = FeatureFlagDelegate()

    delegate.add('a', 'xyz')
    delegate.merge([
        FeatureFlag('a', 'abc'),
        FeatureFlag('b'),
        FeatureFlag('c', '111'),
        FeatureFlag('d'),
    ])

    assert delegate.to_json() == [
        {'featureFlag': 'a', 'variant': 'abc'},
        {'featureFlag': 'b'},
        {'featureFlag': 'c', 'variant': '111'},
        {'featureFlag': 'd'}
    ]


def test_delegate_merge_ignores_anything_that_isnt_a_feature_flag_instance():
    delegate = FeatureFlagDelegate()

    delegate.merge([
        FeatureFlag('a', None),
        1234,
        FeatureFlag('b'),
        'hello',
        FeatureFlag('c', '111'),
        Exception(':)'),
        FeatureFlag('d'),
        None
    ])

    assert delegate.to_json() == [
        {'featureFlag': 'a'},
        {'featureFlag': 'b'},
        {'featureFlag': 'c', 'variant': '111'},
        {'featureFlag': 'd'}
    ]


@pytest.mark.parametrize('invalid_name', _invalid_names)
def test_delegate_merge_drops_flag_when_name_is_invalid(invalid_name):
    delegate = FeatureFlagDelegate()
    delegate.merge([
        FeatureFlag('abc', '123'),
        FeatureFlag(invalid_name, '456'),
        FeatureFlag('xyz', '789')
    ])

    assert delegate.to_json() == [
        {'featureFlag': 'abc', 'variant': '123'},
        {'featureFlag': 'xyz', 'variant': '789'}
    ]


def test_delegate_can_remove_flags_by_name():
    delegate = FeatureFlagDelegate()
    delegate.merge([
        FeatureFlag('abc', '123'),
        FeatureFlag('to be removed', '456'),
        FeatureFlag('xyz', '789')
    ])

    delegate.remove('to be removed')

    assert delegate.to_json() == [
        {'featureFlag': 'abc', 'variant': '123'},
        {'featureFlag': 'xyz', 'variant': '789'}
    ]


def test_delegate_remove_does_nothing_when_no_flag_has_the_given_name():
    delegate = FeatureFlagDelegate()
    delegate.merge([
        FeatureFlag('abc', '123'),
        FeatureFlag('to be kept', '456'),
        FeatureFlag('xyz', '789')
    ])

    delegate.remove('to be removed')

    assert delegate.to_json() == [
        {'featureFlag': 'abc', 'variant': '123'},
        {'featureFlag': 'to be kept', 'variant': '456'},
        {'featureFlag': 'xyz', 'variant': '789'}
    ]


def test_delegate_can_remove_all_flags_at_once():
    delegate = FeatureFlagDelegate()
    delegate.merge([
        FeatureFlag('abc', '123'),
        FeatureFlag('to be kept', '456'),
        FeatureFlag('xyz', '789')
    ])

    delegate.clear()

    assert delegate.to_json() == []


def test_delegate_clear_does_nothing_when_there_are_no_flags():
    delegate = FeatureFlagDelegate()
    delegate.clear()

    assert delegate.to_json() == []


def test_delegate_to_list_returns_a_list_of_feature_flags():
    delegate = FeatureFlagDelegate()
    delegate.merge([FeatureFlag('a'), FeatureFlag('b'), FeatureFlag('c')])

    assert delegate.to_list() == [
        FeatureFlag('a'),
        FeatureFlag('b'),
        FeatureFlag('c')
    ]


def test_delegate_can_be_mutated_without_affecting_the_internal_storage():
    delegate = FeatureFlagDelegate()
    delegate.merge([FeatureFlag('a'), FeatureFlag('b')])

    flags = delegate.to_list()
    flags.pop()
    flags.append(1234)
    flags.append(5678)

    assert flags == [FeatureFlag('a'), 1234, 5678]
    assert delegate.to_list() == [FeatureFlag('a'), FeatureFlag('b')]



import os
import socket
import logging
import random
import sys
import time
import unittest
from pathlib import PurePath, Path
from unittest.mock import patch
from io import StringIO
from threading import Thread

from bugsnag.breadcrumbs import BreadcrumbType
from bugsnag.configuration import Configuration
from bugsnag.error import Error
from bugsnag.middleware import DefaultMiddleware, SimpleMiddleware
from bugsnag.sessiontracker import SessionMiddleware

import pytest


class TestConfiguration(unittest.TestCase):

    def test_reads_api_key_from_environ(self):
        os.environ['BUGSNAG_API_KEY'] = 'aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa'
        c = Configuration()
        self.assertEqual(c.api_key, 'aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa')
        self.assertEqual(c.project_root, os.getcwd())

    def test_should_notify(self):
        # Test custom release_stage
        c = Configuration()
        c.release_stage = "anything"
        self.assertTrue(c.should_notify())

        # Test release_stage in notify_release_stages
        c = Configuration()
        c.notify_release_stages = ["production"]
        c.release_stage = "development"
        self.assertFalse(c.should_notify())

        # Test release_stage in notify_release_stages
        c = Configuration()
        c.notify_release_stages = ["custom"]
        c.release_stage = "custom"
        self.assertTrue(c.should_notify())

    def test_ignore_classes_with_a_single_exception(self):
        # Test ignoring a class works with a single Exception
        c = Configuration()
        c.ignore_classes.append("SystemError")
        self.assertTrue(c.should_ignore(SystemError("Example")))

        c = Configuration()
        c.ignore_classes.append("SystemError")
        self.assertFalse(c.should_ignore(Exception("Example")))

    def test_ignore_classes_works_with_list_of_errors(self):
        # Test ignoring a class works with a list of Errors
        c = Configuration()
        c.ignore_classes.append("SystemError")

        error1 = Error('SystemError', 'ahh', [])
        error2 = Error('AnotherError', 'ahh', [])
        error3 = Error('SomeOtherError', 'ahh', [])
        error4 = Error('Exception', 'ahh', [])

        assert c.should_ignore([error1, error2, error3, error4])
        assert c.should_ignore([error4, error3, error2, error1])
        assert c.should_ignore([error1])

        assert not c.should_ignore([error2, error3, error4])
        assert not c.should_ignore([error4, error3, error2])

    def test_hostname(self):
        c = Configuration()
        self.assertEqual(c.hostname, socket.gethostname())

        os.environ["DYNO"] = "YES"
        c = Configuration()
        self.assertEqual(c.hostname, None)

    def test_session_tracking_defaults(self):
        c = Configuration()
        self.assertTrue(c.auto_capture_sessions)
        self.assertEqual(c.session_endpoint, "https://sessions.bugsnag.com")

    def test_default_middleware_location(self):
        c = Configuration()

        assert len(c.internal_middleware.stack) == 3
        assert isinstance(c.internal_middleware.stack[0], SimpleMiddleware)
        assert c.internal_middleware.stack[1] is DefaultMiddleware
        assert c.internal_middleware.stack[2] is SessionMiddleware

        assert len(c.middleware.stack) == 0

    def test_validate_api_key(self):
        c = Configuration()
        with pytest.raises(TypeError) as e:
            c.configure(api_key=[])
            assert str(e) == 'api_key should be str, got list.'
        c.configure(api_key='ffff')
        assert c.api_key == 'ffff'

    def test_validate_endpoint(self):
        c = Configuration()
        with pytest.raises(TypeError) as e:
            c.configure(endpoint=56)

            assert str(e) == 'endpoint should be str, got int. '
            assert c.endpoint == 'https://notify.bugsnag.com'

        c.configure(endpoint='https://notify.example.com')
        assert c.endpoint == 'https://notify.example.com'

    def test_validate_app_type(self):
        c = Configuration()
        assert c.app_type is None
        with pytest.warns(RuntimeWarning) as record:
            c.configure(app_type=[])

            assert len(record) == 1
            assert str(record[0].message) == 'app_type should be str, got list'
            assert c.app_type is None

            c.configure(app_type='rq')

            assert len(record) == 1
            assert c.app_type == 'rq'

    def test_validate_app_version(self):
        c = Configuration()
        with pytest.warns(RuntimeWarning) as record:
            c.configure(app_version=[])

            assert len(record) == 1
            assert (str(record[0].message) ==
                    'app_version should be str, got list')
            assert c.app_version is None

            c.configure(app_version='1.2.6')

            assert len(record) == 1
            assert c.app_version == '1.2.6'

    def test_validate_asynchronous(self):
        c = Configuration()
        assert c.asynchronous is True
        with pytest.warns(RuntimeWarning) as record:
            c.configure(asynchronous='true')

            assert len(record) == 1
            assert (str(record[0].message) ==
                    'asynchronous should be bool, got str')
            assert c.asynchronous is True

            c.configure(asynchronous=False)
            assert len(record) == 1
            assert c.asynchronous is False

    def test_validate_auto_notify(self):
        c = Configuration()
        assert c.auto_notify is True
        with pytest.warns(RuntimeWarning) as record:
            c.configure(auto_notify='true')

            assert len(record) == 1
            assert (str(record[0].message) ==
                    'auto_notify should be bool, got str')

            c.configure(auto_notify=False)
            assert len(record) == 1
            assert c.auto_notify is False

    def test_validate_auto_capture_sessions(self):
        c = Configuration()
        assert c.auto_capture_sessions is True
        with pytest.warns(RuntimeWarning) as record:
            c.configure(auto_capture_sessions='true')

            assert len(record) == 1
            assert (str(record[0].message) ==
                    'auto_capture_sessions should be bool, got str')

            c.configure(auto_capture_sessions=False)
            assert len(record) == 1
            assert c.auto_capture_sessions is False

    def test_validate_delivery(self):
        c = Configuration()
        assert c.delivery is not None

        class BadDelivery(object):
            def deliv(self, *args, **kwargs):
                pass

        class GoodDelivery(object):
            def deliver(self, *args, **kwargs):
                pass

        with pytest.warns(RuntimeWarning) as record:
            c.configure(delivery=BadDelivery())

            assert len(record) == 1
            assert (str(record[0].message) ==
                    ('delivery should implement Delivery interface, ' +
                     'got BadDelivery. This will be an error in a future ' +
                     'release.'))

            good = GoodDelivery()
            c.configure(delivery=good)
            assert len(record) == 1
            assert c.delivery == good

    def test_validate_hostname(self):
        c = Configuration()
        with pytest.warns(RuntimeWarning) as record:
            c.configure(hostname=[])

            assert len(record) == 1
            assert (str(record[0].message) ==
                    'hostname should be str, got list')
            assert c.hostname is None

            c.configure(hostname='testserver')

            assert len(record) == 1
            assert c.hostname == 'testserver'

    def test_validate_ignore_classes(self):
        c = Configuration()
        with pytest.warns(RuntimeWarning) as record:
            c.configure(ignore_classes=302)

            assert len(record) == 1
            assert (str(record[0].message) ==
                    'ignore_classes should be list or tuple, got int')
            assert isinstance(c.ignore_classes, (list, tuple))

            c.configure(ignore_classes=['LookupError'])

            assert len(record) == 1
            assert c.ignore_classes == ['LookupError']

    def test_validate_lib_root(self):
        c = Configuration()
        with pytest.warns(RuntimeWarning) as record:
            c.configure(lib_root=False)

            assert len(record) == 1
            assert (str(record[0].message) ==
                    'lib_root should be str, got bool')

            c.configure(lib_root='/path/to/python/lib')

            assert len(record) == 1
            assert c.lib_root == '/path/to/python/lib'

    def test_validate_notify_release_stages(self):
        c = Configuration()
        with pytest.warns(RuntimeWarning) as record:
            c.configure(notify_release_stages='beta')

            assert len(record) == 1
            assert (str(record[0].message) ==
                    'notify_release_stages should be list or tuple, got str')
            assert c.notify_release_stages is None

            c.configure(notify_release_stages=['beta'])

            assert len(record) == 1
            assert c.notify_release_stages == ['beta']

    def test_validate_params_filters(self):
        c = Configuration()
        with pytest.warns(RuntimeWarning) as record:
            c.configure(params_filters='pw')

            assert len(record) == 1
            assert (str(record[0].message) ==
                    'params_filters should be list or tuple, got str')
            assert isinstance(c.params_filters, list)

            c.configure(params_filters=['pw'])

            assert len(record) == 1
            assert c.params_filters == ['pw']

    def test_validate_project_root(self):
        c = Configuration()

        if sys.version_info < (3, 6):
            expected_type = 'PurePath'
        else:
            expected_type = 'PathLike'

        with pytest.warns(RuntimeWarning) as record:
            c.configure(project_root=True)

            assert len(record) == 1
            assert str(record[0].message) == \
                'project_root should be str or %s, got bool' % expected_type
            assert c.project_root == os.getcwd()

            c.configure(project_root='/path/to/python/project')

            assert len(record) == 1
            assert c.project_root == '/path/to/python/project'

            c.configure(project_root=Path('/path/to/python/project'))

            assert len(record) == 1
            assert c.project_root == '/path/to/python/project'

            c.configure(project_root=PurePath('/path/to/python/project'))

            assert len(record) == 1
            assert c.project_root == '/path/to/python/project'

    def test_validate_proxy_host(self):
        c = Configuration()
        with pytest.warns(RuntimeWarning) as record:
            c.configure(proxy_host=True)

            assert len(record) == 1
            assert (str(record[0].message) ==
                    'proxy_host should be str, got bool')
            assert c.proxy_host is None

            c.configure(proxy_host='112.0.0.45')

            assert len(record) == 1
            assert c.proxy_host == '112.0.0.45'

    def test_validate_release_stage(self):
        c = Configuration()
        with pytest.warns(RuntimeWarning) as record:
            c.configure(release_stage=False)

            assert len(record) == 1
            assert (str(record[0].message) ==
                    'release_stage should be str, got bool')
            assert c.release_stage == 'production'

            c.configure(release_stage='beta-2')

            assert len(record) == 1
            assert c.release_stage == 'beta-2'

    def test_validate_send_code(self):
        c = Configuration()
        assert c.send_code is True
        with pytest.warns(RuntimeWarning) as record:
            c.configure(send_code='false')

            assert len(record) == 1
            assert (str(record[0].message) ==
                    'send_code should be bool, got str')
            assert c.send_code is True

            c.configure(send_code=False)
            assert len(record) == 1
            assert c.send_code is False

    def test_validate_send_environment(self):
        c = Configuration()
        assert c.send_environment is False
        with pytest.warns(RuntimeWarning) as record:
            c.configure(send_environment='True')

            assert len(record) == 1
            assert (str(record[0].message) ==
                    'send_environment should be bool, got str')
            assert c.send_environment is False

            c.configure(send_environment=True)
            assert len(record) == 1
            assert c.send_environment is True

    def test_validate_session_endpoint(self):
        c = Configuration()
        with pytest.raises(TypeError) as e:
            c.configure(session_endpoint=False)

            assert str(e) == 'session_endpoint should be str, got bool.'
            assert c.session_endpoint == 'https://sessions.bugsnag.com'

        c.configure(session_endpoint='https://sessions.example.com')
        assert c.session_endpoint == 'https://sessions.example.com'

    def test_validate_traceback_exclude_modules(self):
        c = Configuration()
        with pytest.warns(RuntimeWarning) as record:
            c.configure(traceback_exclude_modules='_.logger')

            assert len(record) == 1
            assert (str(record[0].message) ==
                    'traceback_exclude_modules should be list or tuple, ' +
                    'got str')
            assert isinstance(c.traceback_exclude_modules, list)

            c.configure(traceback_exclude_modules=['_.logger'])

            assert len(record) == 1
            assert c.traceback_exclude_modules == ['_.logger']

    def test_validate_unknown_config_option(self):
        c = Configuration()
        with pytest.raises(TypeError):
            c.configure(emperor=True)

    @patch('sys.stderr', new_callable=StringIO)
    def test_default_logger(self, mock_stderr):
        config = Configuration()

        assert isinstance(config.logger, logging.Logger)

        # debug and info calls should be ignored by default
        config.logger.debug('hello debug!')
        config.logger.info('hello info!')
        assert mock_stderr.getvalue() == ''

        config.logger.warning('hello warning!')
        assert mock_stderr.getvalue().endswith(
            '[bugsnag] WARNING - hello warning!\n'
        )

        config.logger.error('hello error!')
        assert mock_stderr.getvalue().endswith(
            '[bugsnag] ERROR - hello error!\n'
        )

        config.logger.critical('hello critical!')
        assert mock_stderr.getvalue().endswith(
            '[bugsnag] CRITICAL - hello critical!\n'
        )

        config.logger.exception('hello exception! %s', Exception('oh no!'))
        assert mock_stderr.getvalue().endswith(
            '[bugsnag] ERROR - hello exception! oh no!\nNoneType: None\n'
        )

    @patch('sys.stderr', new_callable=StringIO)
    def test_default_logger_with_configure_call(self, mock_stderr):
        config = Configuration()
        config.configure(app_type='very good')

        assert isinstance(config.logger, logging.Logger)

        # debug and info calls should be ignored by default
        config.logger.debug('hello debug!')
        config.logger.info('hello info!')
        assert mock_stderr.getvalue() == ''

        config.logger.warning('hello warning!')
        assert mock_stderr.getvalue().endswith(
            '[bugsnag] WARNING - hello warning!\n'
        )

        config.logger.error('hello error!')
        assert mock_stderr.getvalue().endswith(
            '[bugsnag] ERROR - hello error!\n'
        )

        config.logger.critical('hello critical!')
        assert mock_stderr.getvalue().endswith(
            '[bugsnag] CRITICAL - hello critical!\n'
        )

        config.logger.exception('hello exception! %s', Exception('oh no!'))
        assert mock_stderr.getvalue().endswith(
            '[bugsnag] ERROR - hello exception! oh no!\nNoneType: None\n'
        )

    @patch('sys.stderr', new_callable=StringIO)
    def test_logger_can_be_disabled(self, mock_stderr):
        config = Configuration(logger=None)

        assert isinstance(config.logger, logging.Logger)

        config.logger.debug('hello debug!')
        config.logger.info('hello info!')
        config.logger.warning('hello warning!')
        config.logger.error('hello error!')
        config.logger.critical('hello critical!')
        config.logger.exception('hello exception!')

        assert mock_stderr.getvalue() == ''

    @patch('sys.stderr', new_callable=StringIO)
    def test_logger_can_be_disabled_using_configure(self, mock_stderr):
        config = Configuration()
        config.configure(logger=None)

        assert isinstance(config.logger, logging.Logger)

        config.logger.debug('hello debug!')
        config.logger.info('hello info!')
        config.logger.warning('hello warning!')
        config.logger.error('hello error!')
        config.logger.critical('hello critical!')
        config.logger.exception('hello exception!')

        assert mock_stderr.getvalue() == ''

    @patch('sys.stderr', new_callable=StringIO)
    def test_default_logger_is_used_if_invalid_logger_set(self, mock_stderr):
        with pytest.warns(RuntimeWarning) as warnings:
            config = Configuration(logger=123)

        expected_message = 'logger should be logging.Logger, got int'

        assert len(warnings) == 1
        assert warnings[0].message.args[0] == expected_message

        # the default logger should be used
        assert isinstance(config.logger, logging.Logger)

        config.logger.warning('hello warning!')
        assert mock_stderr.getvalue().endswith(
            '[bugsnag] WARNING - hello warning!\n'
        )

    @patch('sys.stderr', new_callable=StringIO)
    def test_default_logger_is_used_if_invalid_logger_set_with_configure(
        self,
        mock_stderr
    ):
        config = Configuration()

        with pytest.warns(RuntimeWarning) as warnings:
            config.configure(logger='hello')

        expected_message = 'logger should be logging.Logger, got str'

        assert len(warnings) == 1
        assert warnings[0].message.args[0] == expected_message

        # the default logger should be used
        assert isinstance(config.logger, logging.Logger)

        config.logger.warning('hello warning!')
        assert mock_stderr.getvalue().endswith(
            '[bugsnag] WARNING - hello warning!\n'
        )

    def test_validate_max_breadcrumbs_type(self):
        c = Configuration()

        with pytest.warns(RuntimeWarning) as record:
            c.configure(max_breadcrumbs='one million')

            assert len(record) == 1

            message = str(record[0].message)
            assert message == 'max_breadcrumbs should be int, got str'
            assert c.max_breadcrumbs == 25

        c.configure(max_breadcrumbs=12)
        assert c.max_breadcrumbs == 12

    def test_validate_max_breadcrumbs_less_than_0(self):
        c = Configuration()

        with pytest.warns(RuntimeWarning) as record:
            c.configure(max_breadcrumbs=-1)

            assert len(record) == 1

            message = str(record[0].message)
            assert message == 'max_breadcrumbs should be an int between 0 ' \
                'and 100, got "-1"'
            assert c.max_breadcrumbs == 25

        c.configure(max_breadcrumbs=0)
        assert c.max_breadcrumbs == 0

    def test_validate_max_breadcrumbs_greater_than_100(self):
        c = Configuration()

        with pytest.warns(RuntimeWarning) as record:
            c.configure(max_breadcrumbs=101)

            assert len(record) == 1

            message = str(record[0].message)
            assert message == 'max_breadcrumbs should be an int between 0 ' \
                'and 100, got "101"'
            assert c.max_breadcrumbs == 25

        c.configure(max_breadcrumbs=100)
        assert c.max_breadcrumbs == 100

    def test_breadcrumb_log_level(self):
        c = Configuration()
        assert c.breadcrumb_log_level == logging.INFO

        c.configure(breadcrumb_log_level=logging.ERROR)
        assert c.breadcrumb_log_level == logging.ERROR

    def test_enabled_breadcrumb_types(self):
        c = Configuration()
        assert c.enabled_breadcrumb_types == list(BreadcrumbType)

        c.configure(enabled_breadcrumb_types=[BreadcrumbType.ERROR])
        assert c.enabled_breadcrumb_types == [BreadcrumbType.ERROR]

    def test_breadcrumbs_are_read_only(self):
        c = Configuration()
        assert c.breadcrumbs == []

        c.breadcrumbs.append('definitely a breadcrumb')
        assert c.breadcrumbs == []

        with pytest.raises(AttributeError) as e:
            c.breadcrumbs = ['??']

            assert str(e) == "AttributeError: can't set attribute"

    def test_on_breadcrumb_callbacks_can_be_added_and_removed(self):
        c = Configuration()
        assert c._on_breadcrumbs == []

        def on_breadcrumb():
            pass

        on_breadcrumb_2 = lambda: None  # noqa: E731

        c.add_on_breadcrumb(on_breadcrumb)
        assert c._on_breadcrumbs == [on_breadcrumb]

        c.add_on_breadcrumb(on_breadcrumb_2)
        assert c._on_breadcrumbs == [on_breadcrumb, on_breadcrumb_2]

        c.remove_on_breadcrumb(on_breadcrumb)
        assert c._on_breadcrumbs == [on_breadcrumb_2]

        c.remove_on_breadcrumb(on_breadcrumb_2)
        assert c._on_breadcrumbs == []

    def test_removing_a_nonexistent_on_breadcrumb_callback_does_nothing(self):
        c = Configuration()

        c.remove_on_breadcrumb(lambda: None)
        assert c._on_breadcrumbs == []

    def test_concurrent_access_to_on_breadcrumb_callbacks(self):
        c = Configuration()

        def first_on_breadcrumb():
            pass

        def second_on_breadcrumb():
            pass

        def add_callbacks():
            try:
                thread.exception = None

                c.add_on_breadcrumb(first_on_breadcrumb)

                # sleep for a bit to allow other threads time to interfere
                time.sleep(random.randrange(0, 100) / 1000)

                c.remove_on_breadcrumb(first_on_breadcrumb)
                c.add_on_breadcrumb(first_on_breadcrumb)

                # sleep for a bit to allow other threads time to interfere
                time.sleep(random.randrange(0, 100) / 1000)

                c.add_on_breadcrumb(second_on_breadcrumb)

                # sleep for a bit to allow other threads time to interfere
                time.sleep(random.randrange(0, 100) / 1000)

                c.remove_on_breadcrumb(second_on_breadcrumb)
                c.add_on_breadcrumb(second_on_breadcrumb)

            except Exception as e:
                thread.exception = e

        number_of_threads = 5

        threads = []
        for i in range(number_of_threads):
            thread = Thread(target=add_callbacks)
            threads.append(thread)

        # shuffle the list of threads so they don't run in a reliable order
        random.shuffle(threads)

        for thread in threads:
            thread.start()

        # shuffle the list of threads so they don't run in a reliable order
        random.shuffle(threads)

        for thread in threads:
            thread.join(2)

            assert not thread.is_alive()

            # if an exception happened in the thread, raise it here instead
            if thread.exception is not None:
                raise thread.exception

        assert len(c._on_breadcrumbs) == number_of_threads * 2
